"""
This module provides a function to easily and efficiently plot trajectories stored in a ragged array.
"""

from clouddrift.ragged import segment, rowsize_to_index
import numpy as np
import pandas as pd
from typing import Optional, Union
import xarray as xr
import pandas as pd
from typing import Optional, Union
from clouddrift.ragged import segment, rowsize_to_index


def plot_ragged(
    ax,
    longitude: Union[list, np.ndarray, pd.Series, xr.DataArray],
    latitude: Union[list, np.ndarray, pd.Series, xr.DataArray],
    rowsize: Union[list, np.ndarray, pd.Series, xr.DataArray],
    *args,
    colors: Optional[Union[list, np.ndarray, pd.Series, xr.DataArray]] = None,
    tolerance: Optional[Union[float, int]] = 180,
    **kwargs,
):
    """Plot trajectories from a ragged array dataset on a Matplotlib Axes
    or a Cartopy GeoAxes object ``ax``.

    This function wraps Matplotlib's ``plot`` function (``plt.plot``) and
    ``LineCollection`` (``matplotlib.collections``) to efficiently plot
    trajectories from a ragged array dataset.

    Parameters
    ----------
    ax: matplotlib.axes.Axes or cartopy.mpl.geoaxes.GeoAxes
        Axis to plot on.
    longitude : array-like
        Longitude sequence. Unidimensional array input.
    latitude : array-like
        Latitude sequence. Unidimensional array input.
    rowsize : list
        List of integers specifying the number of data points in each row.
    *args : tuple
        Additional arguments to pass to ``ax.plot``.
    colors : array-like
        Colors to use for plotting. If colors is the same shape as longitude and latitude,
        the trajectories are splitted into segments and each segment is colored according
        to the corresponding color value. If colors is the same shape as rowsize, the
        trajectories are uniformly colored according to the corresponding color value.
    tolerance : float
        Longitude tolerance gap between data points (in degrees) for segmenting trajectories.
        For periodic domains, the tolerance parameter should be set to the maximum allowed gap
        between data points. Defaults to 180.
    **kwargs : dict
        Additional keyword arguments to pass to ``ax.plot``.

    Returns
    -------
    list of matplotlib.lines.Line2D or matplotlib.collections.LineCollection
        The plotted lines or line collection. Can be used to set a colorbar
        after plotting or extract information from the lines.

    Examples
    --------

    Plot the first 100 trajectories from the gdp1h dataset, assigning
    a different color to each trajectory:

    >>> from clouddrift import datasets
    >>> import matplotlib.pyplot as plt
    >>> ds = datasets.gdp1h()
    >>> ds = subset(ds, {"ID": ds.ID[:100].values}).load()
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(1, 1, 1)

    >>> plot_ragged(
    >>>     ax,
    >>>     ds.lon,
    >>>     ds.lat,
    >>>     ds.rowsize,
    >>>     colors=np.arange(len(ds.rowsize))
    >>> )

    To plot the same trajectories, but assigning a different color to each
    observation and specifying a colormap:

    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(1, 1, 1)
    >>> time = [v.astype(np.int64) / 86400 / 1e9 for v in ds.time.values]
    >>> lc = plot_ragged(
    >>>     ax,
    >>>     ds.lon,
    >>>     ds.lat,
    >>>     ds.rowsize,
    >>>     colors=np.floor(time),
    >>>     cmap="inferno"
    >>> )
    >>> fig.colorbar(lc[0])
    >>> ax.set_xlim([-180, 180])
    >>> ax.set_ylim([-90, 90])

    Finally, to plot the same trajectories, but using a cartopy
    projection:

    >>> import cartopy.crs as ccrs
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(1, 1, 1, projection=ccrs.Mollweide())
    >>> time = [v.astype(np.int64) / 86400 / 1e9 for v in ds.time.values]
    >>> lc = plot_ragged(
    >>>     ax,
    >>>     ds.lon,
    >>>     ds.lat,
    >>>     ds.rowsize,
    >>>     colors=np.arange(len(ds.rowsize)),
    >>>     transform=ccrs.PlateCarree(),
    >>>     cmap=cmocean.cm.ice,
    >>> )

    Raises
    ------
    ValueError
        If longitude and latitude arrays do not have the same shape.
        If colors do not have the same shape as longitude and latitude arrays or rowsize.
        If ax is not a matplotlib Axes or GeoAxes object.
        If ax is a GeoAxes object and the transform keyword argument is not provided.

    ImportError
        If matplotlib is not installed.
        If the axis is a GeoAxes object and cartopy is not installed.
    """

    # optional dependency
    try:
        import matplotlib.pyplot as plt
        import matplotlib.colors as mcolors
        from matplotlib.collections import LineCollection
        from matplotlib import cm
    except ImportError:
        raise ImportError("missing optional dependency 'matplotlib'")

    if hasattr(ax, "coastlines"):  # check if GeoAxes without cartopy
        try:
            from cartopy.mpl.geoaxes import GeoAxes

            if isinstance(ax, GeoAxes) and not kwargs.get("transform"):
                raise ValueError(
                    "For GeoAxes, the transform keyword argument must be provided."
                )
        except ImportError:
            raise ImportError("missing optional dependency 'cartopy'")
    elif not isinstance(ax, plt.Axes):
        raise ValueError("ax must be either: plt.Axes or GeoAxes.")

    if np.sum(rowsize) != len(longitude):
        raise ValueError("The sum of rowsize must equal the length of lon and lat.")

    if len(longitude) != len(latitude):
        raise ValueError("lon and lat must have the same length.")

    if colors is None:
        colors = np.arange(len(rowsize))
    elif colors is not None and (len(colors) not in [len(longitude), len(rowsize)]):
        raise ValueError("shape colors must match the shape of lon/lat or rowsize.")

    # define a colormap
    cmap = kwargs.pop("cmap", cm.viridis)

    # define a normalization obtain uniform colors
    # for the sequence of lines or LineCollection
    norm = kwargs.pop(
        "norm", mcolors.Normalize(vmin=np.nanmin(colors), vmax=np.nanmax(colors))
    )

    mpl_plot = True if colors is None or len(colors) == len(rowsize) else False
    traj_idx = rowsize_to_index(rowsize)

    lines = []
    for i in range(len(rowsize)):
        lon_i, lat_i = (
            longitude[traj_idx[i] : traj_idx[i + 1]],
            latitude[traj_idx[i] : traj_idx[i + 1]],
        )

        start = 0
        for length in segment(lon_i, tolerance, rowsize=segment(lon_i, -tolerance)):
            end = start + length

            if mpl_plot:
                line = ax.plot(
                    lon_i[start:end],
                    lat_i[start:end],
                    c=cmap(norm(colors[i])) if colors is not None else None,
                    *args,
                    **kwargs,
                )
            else:
                colors_i = colors[traj_idx[i] : traj_idx[i + 1]]
                segments = np.column_stack(
                    [
                        lon_i[start : end - 1],
                        lat_i[start : end - 1],
                        lon_i[start + 1 : end],
                        lat_i[start + 1 : end],
                    ]
                ).reshape(-1, 2, 2)
                line = LineCollection(segments, cmap=cmap, norm=norm, *args, **kwargs)
                line.set_array(
                    # color of a segment is the average of its two data points
                    np.convolve(colors_i[start:end], [0.5, 0.5], mode="valid")
                )
                ax.add_collection(line)

            start = end
            lines.append(line)

    # set axis limits
    ax.set_xlim([np.min(longitude), np.max(longitude)])
    ax.set_ylim([np.min(latitude), np.max(latitude)])

    return lines



"""
Functions for kinematic computations.
"""

import numpy as np
import pandas as pd
from typing import Optional, Tuple, Union
import xarray as xr
from clouddrift.sphere import (
    EARTH_RADIUS_METERS,
    bearing,
    cartesian_to_spherical,
    cartesian_to_tangentplane,
    coriolis_frequency,
    distance,
    position_from_distance_and_bearing,
    recast_lon360,
    spherical_to_cartesian,
)
from clouddrift.wavelet import morse_logspace_freq, morse_wavelet, wavelet_transform


def kinetic_energy(
    u: Union[float, list, np.ndarray, xr.DataArray, pd.Series],
    v: Optional[Union[float, list, np.ndarray, xr.DataArray, pd.Series]] = None,
) -> Union[float, np.ndarray, xr.DataArray]:
    """Compute kinetic energy from zonal and meridional velocities.

    Parameters
    ----------
    u : float or array-like
        Zonal velocity.
    v : float or array-like, optional.
        Meridional velocity. If not provided, the flow is assumed one-dimensional
        in time and defined by ``u``.

    Returns
    -------
    ke : float or array-like
        Kinetic energy.

    Examples
    --------
    >>> import numpy as np
    >>> from clouddrift.kinematics import kinetic_energy
    >>> u = np.array([1., 2., 3., 4.])
    >>> v = np.array([1., 1., 1., 1.])
    >>> kinetic_energy(u, v)
    array([1. , 2.5, 5. , 8.5])

    >>> u = np.reshape(np.tile([1., 2., 3., 4.], 2), (2, 4))
    >>> v = np.reshape(np.tile([1., 1., 1., 1.], 2), (2, 4))
    >>> kinetic_energy(u, v)
    array([[1. , 2.5, 5. , 8.5],
           [1. , 2.5, 5. , 8.5]])
    """
    if v is None:
        v = np.zeros_like(u)
    ke = (u**2 + v**2) / 2
    return ke


def inertial_oscillation_from_position(
    longitude: np.ndarray,
    latitude: np.ndarray,
    relative_bandwidth: Optional[float] = None,
    wavelet_duration: Optional[float] = None,
    time_step: Optional[float] = 3600.0,
    relative_vorticity: Optional[Union[float, np.ndarray]] = 0.0,
) -> np.ndarray:
    """Extract inertial oscillations from consecutive geographical positions.

    This function acts by performing a time-frequency analysis of horizontal displacements
    with analytic Morse wavelets. It extracts the portion of the wavelet transform signal
    that follows the inertial frequency (opposite of Coriolis frequency) as a function of time,
    potentially shifted in frequency by a measure of relative vorticity. The result is a pair
    of zonal and meridional relative displacements in meters.

    This function is equivalent to a bandpass filtering of the horizontal displacements. The characteristics
    of the filter are defined by the relative bandwidth of the wavelet transform or by the duration of the wavelet,
    see the parameters below.

    Parameters
    ----------
    longitude : array-like
        Longitude sequence. Unidimensional array input.
    latitude : array-like
        Latitude sequence. Unidimensional array input.
    relative_bandwidth : float, optional
        Bandwidth of the frequency-domain equivalent filter for the extraction of the inertial
        oscillations; a number less or equal to one which is a fraction of the inertial frequency.
        A value of 0.1 leads to a bandpass filter equivalent of +/- 10 percent of the inertial frequency.
    wavelet_duration : float, optional
        Duration of the wavelet, or inverse of the relative bandwidth, which can be passed instead of the
        relative bandwidth.
    time_step : float, optional
        The constant time interval between data points in seconds. Default is 3600.
    relative_vorticity: Optional, float or array-like
        Relative vorticity adding to the local Coriolis frequency. If "f" is the Coriolis
        frequency then "f" + `relative_vorticity` will be the effective Coriolis frequency as defined by Kunze (1985).
        Positive values correspond to cyclonic vorticity, irrespectively of the latitudes of the data
        points.

    Returns
    -------
    xhat : array-like
        Zonal relative displacement in meters from inertial oscillations.
    yhat : array-like
        Meridional relative displacement in meters from inertial oscillations.

    Examples
    --------
    To extract displacements from inertial oscillations from sequences of longitude
    and latitude values, equivalent to bandpass around 20 percent of the local inertial frequency:

    >>> xhat, yhat = inertial_oscillation_from_position(longitude, latitude, relative_bandwidth=0.2)

    The same result can be obtained by specifying the wavelet duration instead of the relative bandwidth:

    >>> xhat, yhat = inertial_oscillation_from_position(longitude, latitude, wavelet_duration=5)

    Next, the residual positions from the inertial displacements can be obtained with another function:

    >>> residual_longitudes, residual_latitudes = residual_position_from_displacement(longitude, latitude, xhat, yhat)

    Raises
    ------
    ValueError
        If longitude and latitude arrays do not have the same shape.
        If relative_vorticity is an array and does not have the same shape as longitude and latitude.
        If time_step is not a float.
        If both relative_bandwidth and wavelet_duration are specified.
        If neither relative_bandwidth nor wavelet_duration are specified.
        If the absolute value of relative_bandwidth is not in the range (0,1].
        If the wavelet duration is not greater than or equal to 1.

    See Also
    --------
    :func:`residual_position_from_displacement`, `wavelet_transform`, `morse_wavelet`

    """
    if longitude.shape != latitude.shape:
        raise ValueError("longitude and latitude arrays must have the same shape.")

    if relative_bandwidth is not None and wavelet_duration is not None:
        raise ValueError(
            "Only one of 'relative_bandwidth' and 'wavelet_duration' can be specified"
        )
    elif relative_bandwidth is None and wavelet_duration is None:
        raise ValueError(
            "One of 'relative_bandwidth' and 'wavelet_duration' must be specified"
        )

    # length of data sequence
    data_length = longitude.shape[0]

    if isinstance(relative_vorticity, float):
        relative_vorticity = np.full_like(longitude, relative_vorticity)
    elif isinstance(relative_vorticity, np.ndarray):
        if not relative_vorticity.shape == longitude.shape:
            raise ValueError(
                "relative_vorticity must be a float or the same shape as longitude and latitude."
            )
    if relative_bandwidth is not None:
        if not 0 < np.abs(relative_bandwidth) <= 1:
            raise ValueError("relative_bandwidth must be in the (0, 1]) range")

    if wavelet_duration is not None:
        if not wavelet_duration >= 1:
            raise ValueError("wavelet_duration must be greater than or equal to 1")

    # wavelet parameters are gamma and beta
    gamma = 3  # symmetric wavelet
    density = 16  # results relative insensitive to this parameter
    # calculate beta from wavelet duration or from relative bandwidth
    if relative_bandwidth is not None:
        wavelet_duration = 1 / np.abs(relative_bandwidth)  # P parameter
    beta = wavelet_duration**2 / gamma

    if isinstance(latitude, xr.DataArray):
        latitude = latitude.to_numpy()
    if isinstance(longitude, xr.DataArray):
        longitude = longitude.to_numpy()

    # Instantaneous absolute frequency of oscillations along trajectory in radian per second
    cor_freq = np.abs(
        coriolis_frequency(latitude) + relative_vorticity * np.sign(latitude)
    )
    cor_freq_max = np.max(cor_freq * 1.05)
    cor_freq_min = np.max(
        [np.min(cor_freq * 0.95), 2 * np.pi / (time_step * data_length)]
    )

    # logarithmically distributed frequencies for wavelet analysis
    radian_frequency = morse_logspace_freq(
        gamma,
        beta,
        data_length,
        (0.05, cor_freq_max * time_step),
        (5, cor_freq_min * time_step),
        density,
    )  # frequencies in radian per unit time

    # wavelet transform on a sphere
    # unwrap longitude recasted in [0,360)
    longitude_unwrapped = np.unwrap(recast_lon360(longitude), period=360)

    # convert lat/lon to Cartesian coordinates x, y , z
    x, y, z = spherical_to_cartesian(longitude_unwrapped, latitude)

    # wavelet transform of x, y, z
    wavelet, _ = morse_wavelet(data_length, gamma, beta, radian_frequency)
    wx = wavelet_transform(x, wavelet, boundary="mirror")
    wy = wavelet_transform(y, wavelet, boundary="mirror")
    wz = wavelet_transform(z, wavelet, boundary="mirror")

    longitude_new, latitude_new = cartesian_to_spherical(
        x - np.real(wx), y - np.real(wy), z - np.real(wz)
    )

    # convert transforms to horizontal displacements on tangent plane
    wxh, wyh = cartesian_to_tangentplane(wx, wy, wz, longitude_new, latitude_new)

    # rotary wavelet transforms to select inertial component; need to divide by sqrt(2)
    wp = (wxh + 1j * wyh) / np.sqrt(2)
    wn = (wxh - 1j * wyh) / np.sqrt(2)

    # find the values of radian_frequency/dt that most closely match cor_freq
    frequency_bins = [
        np.argmin(np.abs(cor_freq[i] - radian_frequency / time_step))
        for i in range(data_length)
    ]

    # get the transform at the inertial and "anti-inertial" frequencies
    # extract the values of wp and wn at the calculated index as a function of time
    # positive is anticyclonic (inertial) in the southern hemisphere
    # negative is anticyclonic (inertial) in the northern hemisphere
    wp = wp[frequency_bins, np.arange(0, data_length)]
    wn = wn[frequency_bins, np.arange(0, data_length)]

    # indices of northern latitude points
    north = latitude >= 0

    # initialize the zonal and meridional components of inertial displacements
    wxhat = np.zeros_like(latitude, dtype=np.complex64)
    wyhat = np.zeros_like(latitude, dtype=np.complex64)
    # equations are x+ = 0.5*(z+ + z-) and y+ = -0.5*1j*(z+ - z-)
    if any(north):
        wxhat[north] = wn[north] / np.sqrt(2)
        wyhat[north] = 1j * wn[north] / np.sqrt(2)
    if any(~north):
        wxhat[~north] = wp[~north] / np.sqrt(2)
        wyhat[~north] = -1j * wp[~north] / np.sqrt(2)

    # inertial displacement in meters
    xhat = np.real(wxhat)
    yhat = np.real(wyhat)

    return xhat, yhat


def residual_position_from_displacement(
    longitude: Union[float, np.ndarray, xr.DataArray],
    latitude: Union[float, np.ndarray, xr.DataArray],
    x: Union[float, np.ndarray],
    y: Union[float, np.ndarray],
) -> Union[Tuple[float], Tuple[np.ndarray]]:
    """
    Return residual longitudes and latitudes along a trajectory on the spherical Earth
    after correcting for zonal and meridional displacements x and y in meters.

    This is applicable as an example when one seeks to correct a trajectory for
    horizontal oscillations due to inertial motions, tides, etc.

    Parameters
    ----------
    longitude : float or array-like
        Longitude in degrees.
    latitude : float or array-like
        Latitude in degrees.
    x : float or np.ndarray
        Zonal displacement in meters.
    y : float or np.ndarray
        Meridional displacement in meters.

    Returns
    -------
    residual_longitude : float or np.ndarray
        Residual longitude after correcting for zonal displacement, in degrees.
    residual_latitude : float or np.ndarray
        Residual latitude after correcting for meridional displacement, in degrees.

    Examples
    --------
    Obtain the new geographical position for a displacement of 1/360-th of the
    circumference of the Earth from original position (longitude,latitude) = (1,0):

    >>> from clouddrift.sphere import EARTH_RADIUS_METERS
    >>> residual_position_from_displacement(1,0,2 * np.pi * EARTH_RADIUS_METERS / 360,0)
    (0.0, 0.0)
    """
    # convert to numpy arrays to insure consistent outputs
    if isinstance(longitude, xr.DataArray):
        longitude = longitude.to_numpy()
    if isinstance(latitude, xr.DataArray):
        latitude = latitude.to_numpy()

    latitudehat = 180 / np.pi * y / EARTH_RADIUS_METERS
    longitudehat = (
        180 / np.pi * x / (EARTH_RADIUS_METERS * np.cos(np.radians(latitude)))
    )

    residual_latitude = latitude - latitudehat
    residual_longitude = recast_lon360(
        np.degrees(np.angle(np.exp(1j * np.radians(longitude - longitudehat))))
    )

    return residual_longitude, residual_latitude


def position_from_velocity(
    u: np.ndarray,
    v: np.ndarray,
    time: np.ndarray,
    x_origin: float,
    y_origin: float,
    coord_system: Optional[str] = "spherical",
    integration_scheme: Optional[str] = "forward",
    time_axis: Optional[int] = -1,
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute positions from arrays of velocities and time and a pair of origin
    coordinates.

    The units of the result are degrees if ``coord_system == "spherical"`` (default).
    If ``coord_system == "cartesian"``, the units of the result are equal to the
    units of the input velocities multiplied by the units of the input time.
    For example, if the input velocities are in meters per second and the input
    time is in seconds, the units of the result will be meters.

    Integration scheme can take one of three values:

        1. "forward" (default): integration from x[i] to x[i+1] is performed
            using the velocity at x[i].
        2. "backward": integration from x[i] to x[i+1] is performed using the
            velocity at x[i+1].
        3. "centered": integration from x[i] to x[i+1] is performed using the
            arithmetic average of the velocities at x[i] and x[i+1]. Note that
            this method introduces some error due to the averaging.

    u, v, and time can be multi-dimensional arrays. If the time axis, along
    which the finite differencing is performed, is not the last one (i.e.
    x.shape[-1]), use the ``time_axis`` optional argument to specify along which
    axis should the differencing be done. ``x``, ``y``, and ``time`` must have
    the same shape.

    This function will not do any special handling of longitude ranges. If the
    integrated trajectory crosses the antimeridian (dateline) in either direction, the
    longitude values will not be adjusted to stay in any specific range such
    as [-180, 180] or [0, 360]. If you need your longitudes to be in a specific
    range, recast the resulting longitude from this function using the function
    :func:`clouddrift.sphere.recast_lon`.

    Parameters
    ----------
    u : np.ndarray
        An array of eastward velocities.
    v : np.ndarray
        An array of northward velocities.
    time : np.ndarray
        An array of time values.
    x_origin : float
        Origin x-coordinate or origin longitude.
    y_origin : float
        Origin y-coordinate or origin latitude.
    coord_system : str, optional
        The coordinate system of the input. Can be "spherical" or "cartesian".
        Default is "spherical".
    integration_scheme : str, optional
        The difference scheme to use for computing the position. Can be
        "forward" or "backward". Default is "forward".
    time_axis : int, optional
        The axis of the time array. Default is -1, which corresponds to the
        last axis.

    Returns
    -------
    x : np.ndarray
        An array of zonal displacements or longitudes.
    y : np.ndarray
        An array of meridional displacements or latitudes.

    Examples
    --------

    Simple integration on a plane, using the forward scheme by default:

    >>> import numpy as np
    >>> from clouddrift.analysis import position_from_velocity
    >>> u = np.array([1., 2., 3., 4.])
    >>> v = np.array([1., 1., 1., 1.])
    >>> time = np.array([0., 1., 2., 3.])
    >>> x, y = position_from_velocity(u, v, time, 0, 0, coord_system="cartesian")
    >>> x
    array([0., 1., 3., 6.])
    >>> y
    array([0., 1., 2., 3.])

    As above, but using centered scheme:

    >>> x, y = position_from_velocity(u, v, time, 0, 0, coord_system="cartesian", integration_scheme="centered")
    >>> x
    array([0., 1.5, 4., 7.5])
    >>> y
    array([0., 1., 2., 3.])

    Simple integration on a sphere (default):

    >>> u = np.array([1., 2., 3., 4.])
    >>> v = np.array([1., 1., 1., 1.])
    >>> time = np.array([0., 1., 2., 3.]) * 1e5
    >>> x, y = position_from_velocity(u, v, time, 0, 0)
    >>> x
    array([0.        , 0.89839411, 2.69584476, 5.39367518])
    >>> y
    array([0.        , 0.89828369, 1.79601515, 2.69201609])

    Integrating across the antimeridian (dateline) by default does not
    recast the resulting longitude:

    >>> u = np.array([1., 1.])
    >>> v = np.array([0., 0.])
    >>> time = np.array([0, 1e5])
    >>> x, y = position_from_velocity(u, v, time, 179.5, 0)
    >>> x
    array([179.5      , 180.3983205])
    >>> y
    array([0., 0.])

    Use the ``clouddrift.sphere.recast_lon`` function to recast the longitudes
    to the desired range:

    >>> from clouddrift.sphere import recast_lon
    >>> recast_lon(x, -180)
    array([ 179.5      , -179.6016795])

    Raises
    ------
    ValueError
        If u and v do not have the same shape.
        If the time axis is outside of the valid range ([-1, N-1]).
        If lengths of x, y, and time along time_axis are not equal.
        If the input coordinate system is not "spherical" or "cartesian".
        If the input integration scheme is not "forward", "backward", or "centered"

    See Also
    --------
    :func:`velocity_from_position`
    """
    # Velocity arrays must have the same shape.
    # Although the exception would be raised further down in the function,
    # we do the check here for a clearer error message.
    if not u.shape == v.shape:
        raise ValueError("u and v must have the same shape.")

    # time_axis must be in valid range
    if time_axis < -1 or time_axis > len(u.shape) - 1:
        raise ValueError(
            f"time_axis ({time_axis}) is outside of the valid range ([-1,"
            f" {len(x.shape) - 1}])."
        )

    # Input arrays must have the same length along the time axis.
    if not u.shape[time_axis] == v.shape[time_axis] == time.shape[time_axis]:
        raise ValueError(
            f"u, v, and time must have the same length along the time axis "
            f"({time_axis})."
        )

    # Swap axes so that we can differentiate along the last axis.
    # This is a syntax convenience rather than memory access optimization:
    # np.swapaxes returns a view of the array, not a copy, if the input is a
    # NumPy array. Otherwise, it returns a copy. For readability, introduce new
    # variable names so that we can more easily differentiate between the
    # original arrays and those with swapped axes.
    u_ = np.swapaxes(u, time_axis, -1)
    v_ = np.swapaxes(v, time_axis, -1)
    time_ = np.swapaxes(time, time_axis, -1)

    x = np.zeros(u_.shape, dtype=u.dtype)
    y = np.zeros(v_.shape, dtype=v.dtype)

    dt = np.diff(time_)

    if integration_scheme.lower() == "forward":
        x[..., 1:] = np.cumsum(u_[..., :-1] * dt, axis=-1)
        y[..., 1:] = np.cumsum(v_[..., :-1] * dt, axis=-1)
    elif integration_scheme.lower() == "backward":
        x[..., 1:] = np.cumsum(u_[1:] * dt, axis=-1)
        y[..., 1:] = np.cumsum(v_[1:] * dt, axis=-1)
    elif integration_scheme.lower() == "centered":
        x[..., 1:] = np.cumsum(0.5 * (u_[..., :-1] + u_[..., 1:]) * dt, axis=-1)
        y[..., 1:] = np.cumsum(0.5 * (v_[..., :-1] + v_[..., 1:]) * dt, axis=-1)
    else:
        raise ValueError(
            'integration_scheme must be "forward", "backward", or "centered".'
        )

    if coord_system.lower() == "cartesian":
        x += x_origin
        y += y_origin
    elif coord_system.lower() == "spherical":
        dx = np.diff(x)
        dy = np.diff(y)
        distances = np.sqrt(dx**2 + dy**2)
        bearings = np.arctan2(dy, dx)
        x[..., 0], y[..., 0] = x_origin, y_origin
        for n in range(distances.shape[-1]):
            x[..., n + 1], y[..., n + 1] = position_from_distance_and_bearing(
                x[..., n], y[..., n], distances[..., n], bearings[..., n]
            )
    else:
        raise ValueError('coord_system must be "spherical" or "cartesian".')

    return np.swapaxes(x, time_axis, -1), np.swapaxes(y, time_axis, -1)


def velocity_from_position(
    x: np.ndarray,
    y: np.ndarray,
    time: np.ndarray,
    coord_system: Optional[str] = "spherical",
    difference_scheme: Optional[str] = "forward",
    time_axis: Optional[int] = -1,
) -> Tuple[xr.DataArray, xr.DataArray]:
    """Compute velocity from arrays of positions and time.

    x and y can be provided as longitude and latitude in degrees if
    coord_system == "spherical" (default), or as easting and northing if
    coord_system == "cartesian".

    The units of the result are meters per unit of time if
    coord_system == "spherical". For example, if the time is provided in the
    units of seconds, the resulting velocity is in the units of meters per
    second. Otherwise, if coord_system == "cartesian", the units of the
    resulting velocity correspond to the units of the input. For example,
    if zonal and meridional displacements are in the units of kilometers and
    time is in the units of hours, the resulting velocity is in the units of
    kilometers per hour.

    x, y, and time can be multi-dimensional arrays. If the time axis, along
    which the finite differencing is performed, is not the last one (i.e.
    x.shape[-1]), use the time_axis optional argument to specify along which
    axis should the differencing be done. x, y, and time must have the same
    shape.

    Difference scheme can take one of three values:

    #. "forward" (default): finite difference is evaluated as ``dx[i] = dx[i+1] - dx[i]``;
    #. "backward": finite difference is evaluated as ``dx[i] = dx[i] - dx[i-1]``;
    #. "centered": finite difference is evaluated as ``dx[i] = (dx[i+1] - dx[i-1]) / 2``.

    Forward and backward schemes are effectively the same except that the
    position at which the velocity is evaluated is shifted one element down in
    the backward scheme relative to the forward scheme. In the case of a
    forward or backward difference scheme, the last or first element of the
    velocity, respectively, is extrapolated from its neighboring point. In the
    case of a centered difference scheme, the start and end boundary points are
    evaluated using the forward and backward difference scheme, respectively.

    Parameters
    ----------
    x : array_like
        An N-d array of x-positions (longitude in degrees or zonal displacement in any unit)
    y : array_like
        An N-d array of y-positions (latitude in degrees or meridional displacement in any unit)
    time : array_like
        An N-d array of times as floating point values (in any unit)
    coord_system : str, optional
        Coordinate system that x and y arrays are in; possible values are "spherical" (default) or "cartesian".
    difference_scheme : str, optional
        Difference scheme to use; possible values are "forward", "backward", and "centered".
    time_axis : int, optional
        Axis along which to differentiate (default is -1)

    Returns
    -------
    u : np.ndarray
        Zonal velocity
    v : np.ndarray
        Meridional velocity

    Raises
    ------
    ValueError
        If x and y do not have the same shape.
        If time_axis is outside of the valid range.
        If lengths of x, y, and time along time_axis are not equal.
        If coord_system is not "spherical" or "cartesian".
        If difference_scheme is not "forward", "backward", or "centered".

    Examples
    --------
    Simple integration on a sphere, using the forward scheme by default:

    >>> import numpy as np
    >>> from clouddrift.kinematics import velocity_from_position
    >>> lon = np.array([0., 1., 3., 6.])
    >>> lat = np.array([0., 1., 2., 3.])
    >>> time = np.array([0., 1., 2., 3.]) * 1e5
    >>> u, v = velocity_from_position(lon, lat, time)
    >>> u
    array([1.11307541, 2.22513331, 3.33515501, 3.33515501])
    >>> v
    array([1.11324496, 1.11409224, 1.1167442 , 1.1167442 ])

    Integration on a Cartesian plane, using the forward scheme by default:

    >>> x = np.array([0., 1., 3., 6.])
    >>> y = np.array([0., 1., 2., 3.])
    >>> time = np.array([0., 1., 2., 3.])
    >>> u, v = velocity_from_position(x, y, time, coord_system="cartesian")
    >>> u
    array([1., 2., 3., 3.])
    >>> v
    array([1., 1., 1., 1.])

    See Also
    --------
    :func:`position_from_velocity`
    """

    # Position arrays must have the same shape.
    # Although the exception would be raised further down in the function,
    # we do the check here for a clearer error message.
    if not x.shape == y.shape:
        raise ValueError("x and y arrays must have the same shape.")

    # time_axis must be in valid range
    if time_axis < -1 or time_axis > len(x.shape) - 1:
        raise ValueError(
            f"time_axis ({time_axis}) is outside of the valid range ([-1,"
            f" {len(x.shape) - 1}])."
        )

    # Input arrays must have the same length along the time axis.
    if not x.shape[time_axis] == y.shape[time_axis] == time.shape[time_axis]:
        raise ValueError(
            f"x, y, and time must have the same length along the time axis "
            f"({time_axis})."
        )

    # Swap axes so that we can differentiate along the last axis.
    # This is a syntax convenience rather than memory access optimization:
    # np.swapaxes returns a view of the array, not a copy, if the input is a
    # NumPy array. Otherwise, it returns a copy. For readability, introduce new
    # variable names so that we can more easily differentiate between the
    # original arrays and those with swapped axes.
    x_ = np.swapaxes(x, time_axis, -1)
    y_ = np.swapaxes(y, time_axis, -1)
    time_ = np.swapaxes(time, time_axis, -1)

    dx = np.empty(x_.shape)
    dy = np.empty(y_.shape)
    dt = np.empty(time_.shape)

    # Compute dx, dy, and dt
    if difference_scheme == "forward":
        # All values except the ending boundary value are computed using the
        # 1st order forward differencing. The ending boundary value is
        # computed using the 1st order backward difference.

        # Time
        dt[..., :-1] = np.diff(time_)
        dt[..., -1] = dt[..., -2]

        # Space
        if coord_system == "cartesian":
            dx[..., :-1] = np.diff(x_)
            dx[..., -1] = dx[..., -2]
            dy[..., :-1] = np.diff(y_)
            dy[..., -1] = dy[..., -2]

        elif coord_system == "spherical":
            distances = distance(x_[..., :-1], y_[..., :-1], x_[..., 1:], y_[..., 1:])
            bearings = bearing(x_[..., :-1], y_[..., :-1], x_[..., 1:], y_[..., 1:])
            dx[..., :-1] = distances * np.cos(bearings)
            dx[..., -1] = dx[..., -2]
            dy[..., :-1] = distances * np.sin(bearings)
            dy[..., -1] = dy[..., -2]

        else:
            raise ValueError('coord_system must be "spherical" or "cartesian".')

    elif difference_scheme == "backward":
        # All values except the starting boundary value are computed using the
        # 1st order backward differencing. The starting boundary value is
        # computed using the 1st order forward difference.

        # Time
        dt[..., 1:] = np.diff(time_)
        dt[..., 0] = dt[..., 1]

        # Space
        if coord_system == "cartesian":
            dx[..., 1:] = np.diff(x_)
            dx[..., 0] = dx[..., 1]
            dy[..., 1:] = np.diff(y_)
            dy[..., 0] = dy[..., 1]

        elif coord_system == "spherical":
            distances = distance(x_[..., :-1], y_[..., :-1], x_[..., 1:], y_[..., 1:])
            bearings = bearing(x_[..., :-1], y_[..., :-1], x_[..., 1:], y_[..., 1:])
            dx[..., 1:] = distances * np.cos(bearings)
            dx[..., 0] = dx[..., 1]
            dy[..., 1:] = distances * np.sin(bearings)
            dy[..., 0] = dy[..., 1]

        else:
            raise ValueError('coord_system must be "spherical" or "cartesian".')

    elif difference_scheme == "centered":
        # Inner values are computed using the 2nd order centered differencing.
        # The start and end boundary values are computed using the 1st order
        # forward and backward differencing, respectively.

        # Time
        dt[..., 1:-1] = (time_[..., 2:] - time_[..., :-2]) / 2
        dt[..., 0] = time_[..., 1] - time_[..., 0]
        dt[..., -1] = time_[..., -1] - time_[..., -2]

        # Space
        if coord_system == "cartesian":
            dx[..., 1:-1] = (x_[..., 2:] - x_[..., :-2]) / 2
            dx[..., 0] = x_[..., 1] - x_[..., 0]
            dx[..., -1] = x_[..., -1] - x_[..., -2]
            dy[..., 1:-1] = (y_[..., 2:] - y_[..., :-2]) / 2
            dy[..., 0] = y_[..., 1] - y_[..., 0]
            dy[..., -1] = y_[..., -1] - y_[..., -2]

        elif coord_system == "spherical":
            # Inner values
            y1 = (y_[..., :-2] + y_[..., 1:-1]) / 2
            x1 = (x_[..., :-2] + x_[..., 1:-1]) / 2
            y2 = (y_[..., 2:] + y_[..., 1:-1]) / 2
            x2 = (x_[..., 2:] + x_[..., 1:-1]) / 2
            distances = distance(x1, y1, x2, y2)
            bearings = bearing(x1, y1, x2, y2)
            dx[..., 1:-1] = distances * np.cos(bearings)
            dy[..., 1:-1] = distances * np.sin(bearings)

            # Boundary values
            distance1 = distance(x_[..., 0], y_[..., 0], x_[..., 1], y_[..., 1])
            bearing1 = bearing(x_[..., 0], y_[..., 0], x_[..., 1], y_[..., 1])
            dx[..., 0] = distance1 * np.cos(bearing1)
            dy[..., 0] = distance1 * np.sin(bearing1)
            distance2 = distance(x_[..., -2], y_[..., -2], x_[..., -1], y_[..., -1])
            bearing2 = bearing(x_[..., -2], y_[..., -2], x_[..., -1], y_[..., -1])
            dx[..., -1] = distance2 * np.cos(bearing2)
            dy[..., -1] = distance2 * np.sin(bearing2)

        else:
            raise ValueError('coord_system must be "spherical" or "cartesian".')

    else:
        raise ValueError(
            'difference_scheme must be "forward", "backward", or "centered".'
        )

    # This should avoid an array copy when returning the result
    dx /= dt
    dy /= dt

    return np.swapaxes(dx, time_axis, -1), np.swapaxes(dy, time_axis, -1)


def spin(
    u: np.ndarray,
    v: np.ndarray,
    time: np.ndarray,
    difference_scheme: Optional[str] = "forward",
    time_axis: Optional[int] = -1,
) -> Union[float, np.ndarray]:
    """Compute spin continuously from velocities and times.

    Spin is traditionally (Sawford, 1999; Veneziani et al., 2005) defined as
    (<u'dv' - v'du'>) / (2 dt EKE) where u' and v' are eddy-perturbations of the
    velocity field, EKE is eddy kinetic energy, dt is the time step, and du' and
    dv' are velocity component increments during dt, and < > denotes ensemble
    average.

    To allow computing spin based on full velocity fields, this function does
    not do any demeaning of the velocity fields. If you need the spin based on
    velocity anomalies, ensure to demean the velocity fields before passing
    them to this function. This function also returns instantaneous spin values,
    so the rank of the result is not reduced relative to the input.

    ``u``, ``v``, and ``time`` can be multi-dimensional arrays. If the time
    axis, along which the finite differencing is performed, is not the last one
    (i.e. ``u.shape[-1]``), use the time_axis optional argument to specify along
    which the spin should be calculated. u, v, and time must either have the
    same shape, or time must be a 1-d array with the same length as
    ``u.shape[time_axis]``.

    Difference scheme can be one of three values:

        1. "forward" (default): finite difference is evaluated as ``dx[i] = dx[i+1] - dx[i]``;
        2. "backward": finite difference is evaluated as ``dx[i] = dx[i] - dx[i-1]``;
        3. "centered": finite difference is evaluated as ``dx[i] = (dx[i+1] - dx[i-1]) / 2``.

    Forward and backward schemes are effectively the same except that the
    position at which the velocity is evaluated is shifted one element down in
    the backward scheme relative to the forward scheme. In the case of a
    forward or backward difference scheme, the last or first element of the
    velocity, respectively, is extrapolated from its neighboring point. In the
    case of a centered difference scheme, the start and end boundary points are
    evaluated using the forward and backward difference scheme, respectively.

    Parameters
    ----------
    u : np.ndarray
        Zonal velocity
    v : np.ndarray
        Meridional velocity
    time : array-like
        Time
    difference_scheme : str, optional
        Difference scheme to use; possible values are "forward", "backward", and "centered".
    time_axis : int, optional
        Axis along which the time varies (default is -1)

    Returns
    -------
    s : float or np.ndarray
        Spin

    Raises
    ------
    ValueError
        If u and v do not have the same shape.
        If the time axis is outside of the valid range ([-1, N-1]).
        If lengths of u, v, and time along time_axis are not equal.
        If difference_scheme is not "forward", "backward", or "centered".

    Examples
    --------
    >>> from clouddrift.kinematics import spin
    >>> import numpy as np
    >>> u = np.array([1., 2., -1., 4.])
    >>> v = np.array([1., 3., -2., 1.])
    >>> time = np.array([0., 1., 2., 3.])
    >>> spin(u, v, time)
    array([ 0.5       , -0.07692308,  1.4       ,  0.41176471])

    Use ``difference_scheme`` to specify an alternative finite difference
    scheme for the velocity differences:

    >>> spin(u, v, time, difference_scheme="centered")
    array([0.5       , 0.        , 0.6       , 0.41176471])
    >>> spin(u, v, time, difference_scheme="backward")
    array([ 0.5       ,  0.07692308, -0.2       ,  0.41176471])

    References
    ----------
    * Sawford, B.L., 1999. Rotation of trajectories in Lagrangian stochastic models of turbulent dispersion. Boundary-layer meteorology, 93, pp.411-424. https://doi.org/10.1023/A:1002114132715
    * Veneziani, M., Griffa, A., Garraffo, Z.D. and Chassignet, E.P., 2005. Lagrangian spin parameter and coherent structures from trajectories released in a high-resolution ocean model. Journal of Marine Research, 63(4), pp.753-788. https://elischolar.library.yale.edu/journal_of_marine_research/100/
    """
    if not u.shape == v.shape:
        raise ValueError("u and v arrays must have the same shape.")

    if not time.shape == u.shape:
        if not time.size == u.shape[time_axis]:
            raise ValueError("time must have the same length as u along time_axis.")

    # axis must be in valid range
    if time_axis < -1 or time_axis > len(u.shape) - 1:
        raise ValueError(
            f"axis ({time_axis}) is outside of the valid range ([-1,"
            f" {len(u.shape) - 1}])."
        )

    # Swap axes so that we can differentiate along the last axis.
    # This is a syntax convenience rather than memory access optimization:
    # np.swapaxes returns a view of the array, not a copy, if the input is a
    # NumPy array. Otherwise, it returns a copy.
    u = np.swapaxes(u, time_axis, -1)
    v = np.swapaxes(v, time_axis, -1)
    time = np.swapaxes(time, time_axis, -1)

    if not time.shape == u.shape:
        # time is 1-d array; broadcast to u.shape.
        time = np.broadcast_to(time, u.shape)

    du = np.empty(u.shape)
    dv = np.empty(v.shape)
    dt = np.empty(time.shape)

    if difference_scheme == "forward":
        du[..., :-1] = np.diff(u)
        du[..., -1] = du[..., -2]
        dv[..., :-1] = np.diff(v)
        dv[..., -1] = dv[..., -2]
        dt[..., :-1] = np.diff(time)
        dt[..., -1] = dt[..., -2]
    elif difference_scheme == "backward":
        du[..., 1:] = np.diff(u)
        du[..., 0] = du[..., 1]
        dv[..., 1:] = np.diff(v)
        dv[..., 0] = dv[..., 1]
        dt[..., 1:] = np.diff(time)
        dt[..., 0] = dt[..., 1]
    elif difference_scheme == "centered":
        du[..., 1:-1] = (u[..., 2:] - u[..., :-2]) / 2
        du[..., 0] = u[..., 1] - u[..., 0]
        du[..., -1] = u[..., -1] - u[..., -2]
        dv[..., 1:-1] = (v[..., 2:] - v[..., :-2]) / 2
        dv[..., 0] = v[..., 1] - v[..., 0]
        dv[..., -1] = v[..., -1] - v[..., -2]
        dt[..., 1:-1] = (time[..., 2:] - time[..., :-2]) / 2
        dt[..., 0] = time[..., 1] - time[..., 0]
        dt[..., -1] = time[..., -1] - time[..., -2]
    else:
        raise ValueError(
            'difference_scheme must be "forward", "backward", or "centered".'
        )

    # Compute spin
    s = (u * dv - v * du) / (2 * dt * kinetic_energy(u, v))

    return np.swapaxes(s, time_axis, -1)



"""
This module provides functions and metadata that can be used to convert the
hourly Global Drifter Program (GDP) data to a ``clouddrift.RaggedArray``
instance.
"""

import clouddrift.adapters.gdp as gdp
from clouddrift.raggedarray import RaggedArray
from clouddrift.adapters.utils import download_with_progress
from datetime import datetime, timedelta
import numpy as np
import urllib.request
import re
import tempfile
from typing import Optional
import os
import warnings
import xarray as xr

GDP_VERSION = "2.01"

GDP_DATA_URL = "https://www.aoml.noaa.gov/ftp/pub/phod/buoydata/hourly_product/v2.01/"
GDP_DATA_URL_EXPERIMENTAL = (
    "https://www.aoml.noaa.gov/ftp/pub/phod/lumpkin/hourly/experimental/"
)
GDP_TMP_PATH = os.path.join(tempfile.gettempdir(), "clouddrift", "gdp")
GDP_TMP_PATH_EXPERIMENTAL = os.path.join(tempfile.gettempdir(), "clouddrift", "gdp_exp")
GDP_DATA = [
    "lon",
    "lat",
    "ve",
    "vn",
    "err_lat",
    "err_lon",
    "err_ve",
    "err_vn",
    "gap",
    "sst",
    "sst1",
    "sst2",
    "err_sst",
    "err_sst1",
    "err_sst2",
    "flg_sst",
    "flg_sst1",
    "flg_sst2",
    "drogue_status",
]


def download(
    drifter_ids: list = None,
    n_random_id: int = None,
    url: str = GDP_DATA_URL,
    tmp_path: str = None,
):
    """Download individual NetCDF files from the AOML server.

    Parameters
    ----------
    drifter_ids : list
        List of drifter to retrieve (Default: all)
    n_random_id : int
        Randomly select n_random_id drifter IDs to download (Default: None)
    url : str
        URL from which to download the data (Default: GDP_DATA_URL). Alternatively, it can be GDP_DATA_URL_EXPERIMENTAL.
    tmp_path : str, optional
        Path to the directory where the individual NetCDF files are stored
        (default varies depending on operating system; /tmp/clouddrift/gdp on Linux)

    Returns
    -------
    out : list
        List of retrieved drifters
    """

    # adjust the tmp_path if using the experimental source
    if tmp_path is None:
        tmp_path = GDP_TMP_PATH if url == GDP_DATA_URL else GDP_TMP_PATH_EXPERIMENTAL

    print(f"Downloading GDP hourly data from {url} to {tmp_path}...")

    # Create a temporary directory if doesn't already exists.
    os.makedirs(tmp_path, exist_ok=True)

    if url == GDP_DATA_URL:
        pattern = "drifter_hourly_[0-9]*.nc"
        filename_pattern = "drifter_hourly_{id}.nc"
    elif url == GDP_DATA_URL_EXPERIMENTAL:
        pattern = "drifter_hourly_[0-9]*.nc"
        filename_pattern = "drifter_hourly_{id}.nc"

    # retrieve all drifter ID numbers
    if drifter_ids is None:
        urlpath = urllib.request.urlopen(url)
        string = urlpath.read().decode("utf-8")
        filelist = re.compile(pattern).findall(string)
        drifter_ids = np.unique([int(f.split("_")[-1][:-3]) for f in filelist])

    # retrieve only a subset of n_random_id trajectories
    if n_random_id:
        if n_random_id > len(drifter_ids):
            warnings.warn(
                f"Retrieving all listed trajectories because {n_random_id} is larger than the {len(drifter_ids)} listed trajectories."
            )
        else:
            rng = np.random.RandomState(42)
            drifter_ids = sorted(rng.choice(drifter_ids, n_random_id, replace=False))

    download_requests = [
        (os.path.join(url, file_name), os.path.join(tmp_path, file_name))
        for file_name in map(lambda d_id: filename_pattern.format(id=d_id), drifter_ids)
    ]
    download_with_progress(download_requests)
    # Download the metadata so we can order the drifter IDs by end date.
    gdp_metadata = gdp.get_gdp_metadata()

    return gdp.order_by_date(gdp_metadata, drifter_ids)


def preprocess(index: int, **kwargs) -> xr.Dataset:
    """Extract and preprocess the Lagrangian data and attributes.

    This function takes an identification number that can be used to create a
    file or url pattern or select data from a Dataframe. It then preprocesses
    the data and returns a clean Xarray Dataset.

    Parameters
    ----------
    index : int
        Drifter's identification number

    Returns
    -------
    ds : xr.Dataset
        Xarray Dataset containing the data and attributes
    """
    ds = xr.load_dataset(
        os.path.join(kwargs["tmp_path"], kwargs["filename_pattern"].format(id=index)),
        decode_times=False,
        decode_coords=False,
    )

    # parse the date with custom function
    ds["deploy_date"].data = gdp.decode_date(np.array([ds.deploy_date.data[0]]))
    ds["end_date"].data = gdp.decode_date(np.array([ds.end_date.data[0]]))
    ds["drogue_lost_date"].data = gdp.decode_date(
        np.array([ds.drogue_lost_date.data[0]])
    )
    ds["time"].data = gdp.decode_date(np.array([ds.time.data[0]]))

    # convert fill values to nan
    for var in [
        "err_lon",
        "err_lat",
        "err_ve",
        "err_vn",
        "sst",
        "sst1",
        "sst2",
        "err_sst",
        "err_sst1",
        "err_sst2",
    ]:
        try:
            ds[var].data = gdp.fill_values(ds[var].data)
        except KeyError:
            warnings.warn(f"Variable {var} not found; skipping.")

    # fix missing values stored as str
    for var in [
        "longitude",
        "latitude",
        "err_lat",
        "err_lon",
        "ve",
        "vn",
        "err_ve",
        "err_vn",
        "sst",
        "sst1",
        "sst2",
    ]:
        try:
            ds[var].encoding["missing value"] = -1e-34
        except KeyError:
            warnings.warn(f"Variable {var} not found in upstream data; skipping.")

    # convert type of some variable
    target_dtype = {
        "ID": "int64",
        "WMO": "int32",
        "expno": "int32",
        "typedeath": "int8",
        "flg_sst": "int8",
        "flg_sst1": "int8",
        "flg_sst2": "int8",
    }

    for var in target_dtype.keys():
        if var in ds.keys():
            ds[var].data = ds[var].data.astype(target_dtype[var])
        else:
            warnings.warn(f"Variable {var} not found in upstream data; skipping.")

    # new variables
    ds["ids"] = (["traj", "obs"], [np.repeat(ds.ID.values, ds.sizes["obs"])])
    ds["drogue_status"] = (
        ["traj", "obs"],
        [gdp.drogue_presence(ds.drogue_lost_date.data, ds.time.data[0])],
    )

    # convert attributes to variable
    ds["location_type"] = (
        ("traj"),
        [False if ds.get("location_type") == "Argos" else True],
    )  # 0 for Argos, 1 for GPS
    ds["DeployingShip"] = (("traj"), gdp.cut_str(ds.DeployingShip, 20))
    ds["DeploymentStatus"] = (("traj"), gdp.cut_str(ds.DeploymentStatus, 20))
    ds["BuoyTypeManufacturer"] = (("traj"), gdp.cut_str(ds.BuoyTypeManufacturer, 20))
    ds["BuoyTypeSensorArray"] = (("traj"), gdp.cut_str(ds.BuoyTypeSensorArray, 20))
    ds["CurrentProgram"] = (
        ("traj"),
        np.int32([gdp.str_to_float(ds.CurrentProgram, -1)]),
    )
    ds["PurchaserFunding"] = (("traj"), gdp.cut_str(ds.PurchaserFunding, 20))
    ds["SensorUpgrade"] = (("traj"), gdp.cut_str(ds.SensorUpgrade, 20))
    ds["Transmissions"] = (("traj"), gdp.cut_str(ds.Transmissions, 20))
    ds["DeployingCountry"] = (("traj"), gdp.cut_str(ds.DeployingCountry, 20))
    ds["DeploymentComments"] = (
        ("traj"),
        gdp.cut_str(
            ds.DeploymentComments.encode("ascii", "ignore").decode("ascii"), 20
        ),
    )  # remove non ascii char
    ds["ManufactureYear"] = (
        ("traj"),
        np.int16([gdp.str_to_float(ds.ManufactureYear, -1)]),
    )
    ds["ManufactureMonth"] = (
        ("traj"),
        np.int16([gdp.str_to_float(ds.ManufactureMonth, -1)]),
    )
    ds["ManufactureSensorType"] = (("traj"), gdp.cut_str(ds.ManufactureSensorType, 20))
    ds["ManufactureVoltage"] = (
        ("traj"),
        np.int16([gdp.str_to_float(ds.ManufactureVoltage[:-6], -1)]),
    )  # e.g. 56 V
    ds["FloatDiameter"] = (
        ("traj"),
        [gdp.str_to_float(ds.FloatDiameter[:-3])],
    )  # e.g. 35.5 cm
    ds["SubsfcFloatPresence"] = (
        ("traj"),
        np.array([gdp.str_to_float(ds.SubsfcFloatPresence)], dtype="bool"),
    )
    ds["DrogueType"] = (("traj"), gdp.cut_str(ds.DrogueType, 7))
    ds["DrogueLength"] = (
        ("traj"),
        [gdp.str_to_float(ds.DrogueLength[:-2])],
    )  # e.g. 4.8 m
    ds["DrogueBallast"] = (
        ("traj"),
        [gdp.str_to_float(ds.DrogueBallast[:-3])],
    )  # e.g. 1.4 kg
    ds["DragAreaAboveDrogue"] = (
        ("traj"),
        [gdp.str_to_float(ds.DragAreaAboveDrogue[:-4])],
    )  # 10.66 m^2
    ds["DragAreaOfDrogue"] = (
        ("traj"),
        [gdp.str_to_float(ds.DragAreaOfDrogue[:-4])],
    )  # e.g. 416.6 m^2
    ds["DragAreaRatio"] = (("traj"), [gdp.str_to_float(ds.DragAreaRatio)])  # e.g. 39.08
    ds["DrogueCenterDepth"] = (
        ("traj"),
        [gdp.str_to_float(ds.DrogueCenterDepth[:-2])],
    )  # e.g. 20.0 m
    ds["DrogueDetectSensor"] = (("traj"), gdp.cut_str(ds.DrogueDetectSensor, 20))

    # vars attributes
    vars_attrs = {
        "ID": {"long_name": "Global Drifter Program Buoy ID", "units": "-"},
        "longitude": {"long_name": "Longitude", "units": "degrees_east"},
        "latitude": {"long_name": "Latitude", "units": "degrees_north"},
        "time": {"long_name": "Time", "units": "seconds since 1970-01-01 00:00:00"},
        "ids": {
            "long_name": "Global Drifter Program Buoy ID repeated along observations",
            "units": "-",
        },
        "rowsize": {
            "long_name": "Number of observations per trajectory",
            "sample_dimension": "obs",
            "units": "-",
        },
        "location_type": {
            "long_name": "Satellite-based location system",
            "units": "-",
            "comments": "0 (Argos), 1 (GPS)",
        },
        "WMO": {
            "long_name": "World Meteorological Organization buoy identification number",
            "units": "-",
        },
        "expno": {"long_name": "Experiment number", "units": "-"},
        "deploy_date": {
            "long_name": "Deployment date and time",
            "units": "seconds since 1970-01-01 00:00:00",
        },
        "deploy_lon": {"long_name": "Deployment longitude", "units": "degrees_east"},
        "deploy_lat": {"long_name": "Deployment latitude", "units": "degrees_north"},
        "start_date": {
            "long_name": "First good date and time derived by DAC quality control",
            "units": "seconds since 1970-01-01 00:00:00",
        },
        "start_lon": {
            "long_name": "First good longitude derived by DAC quality control",
            "units": "degrees_east",
        },
        "start_lat": {
            "long_name": "Last good latitude derived by DAC quality control",
            "units": "degrees_north",
        },
        "end_date": {
            "long_name": "Last good date and time derived by DAC quality control",
            "units": "seconds since 1970-01-01 00:00:00",
        },
        "end_lon": {
            "long_name": "Last good longitude derived by DAC quality control",
            "units": "degrees_east",
        },
        "end_lat": {
            "long_name": "Last good latitude derived by DAC quality control",
            "units": "degrees_north",
        },
        "drogue_lost_date": {
            "long_name": "Date and time of drogue loss",
            "units": "seconds since 1970-01-01 00:00:00",
        },
        "typedeath": {
            "long_name": "Type of death",
            "units": "-",
            "comments": "0 (buoy still alive), 1 (buoy ran aground), 2 (picked up by vessel), 3 (stop transmitting), 4 (sporadic transmissions), 5 (bad batteries), 6 (inactive status)",
        },
        "typebuoy": {
            "long_name": "Buoy type (see https://www.aoml.noaa.gov/phod/dac/dirall.html)",
            "units": "-",
        },
        "DeployingShip": {"long_name": "Name of deployment ship", "units": "-"},
        "DeploymentStatus": {"long_name": "Deployment status", "units": "-"},
        "BuoyTypeManufacturer": {"long_name": "Buoy type manufacturer", "units": "-"},
        "BuoyTypeSensorArray": {"long_name": "Buoy type sensor array", "units": "-"},
        "CurrentProgram": {
            "long_name": "Current Program",
            "units": "-",
            "_FillValue": "-1",
        },
        "PurchaserFunding": {"long_name": "Purchaser funding", "units": "-"},
        "SensorUpgrade": {"long_name": "Sensor upgrade", "units": "-"},
        "Transmissions": {"long_name": "Transmissions", "units": "-"},
        "DeployingCountry": {"long_name": "Deploying country", "units": "-"},
        "DeploymentComments": {"long_name": "Deployment comments", "units": "-"},
        "ManufactureYear": {
            "long_name": "Manufacture year",
            "units": "-",
            "_FillValue": "-1",
        },
        "ManufactureMonth": {
            "long_name": "Manufacture month",
            "units": "-",
            "_FillValue": "-1",
        },
        "ManufactureSensorType": {"long_name": "Manufacture Sensor Type", "units": "-"},
        "ManufactureVoltage": {
            "long_name": "Manufacture voltage",
            "units": "V",
            "_FillValue": "-1",
        },
        "FloatDiameter": {"long_name": "Diameter of surface floater", "units": "cm"},
        "SubsfcFloatPresence": {"long_name": "Subsurface Float Presence", "units": "-"},
        "DrogueType": {"drogue_type": "Drogue Type", "units": "-"},
        "DrogueLength": {"long_name": "Length of drogue.", "units": "m"},
        "DrogueBallast": {
            "long_name": "Weight of the drogue's ballast.",
            "units": "kg",
        },
        "DragAreaAboveDrogue": {"long_name": "Drag area above drogue.", "units": "m^2"},
        "DragAreaOfDrogue": {"long_name": "Drag area drogue.", "units": "m^2"},
        "DragAreaRatio": {"long_name": "Drag area ratio", "units": "m"},
        "DrogueCenterDepth": {
            "long_name": "Average depth of the drogue.",
            "units": "m",
        },
        "DrogueDetectSensor": {"long_name": "Drogue detection sensor", "units": "-"},
        "ve": {"long_name": "Eastward velocity", "units": "m/s"},
        "vn": {"long_name": "Northward velocity", "units": "m/s"},
        "gap": {
            "long_name": "Time interval between previous and next location",
            "units": "s",
        },
        "err_lat": {
            "long_name": "95% confidence interval in latitude",
            "units": "degrees_north",
        },
        "err_lon": {
            "long_name": "95% confidence interval in longitude",
            "units": "degrees_east",
        },
        "err_ve": {
            "long_name": "95% confidence interval in eastward velocity",
            "units": "m/s",
        },
        "err_vn": {
            "long_name": "95% confidence interval in northward velocity",
            "units": "m/s",
        },
        "drogue_status": {
            "long_name": "Status indicating the presence of the drogue",
            "units": "-",
            "flag_values": "1,0",
            "flag_meanings": "drogued, undrogued",
        },
        "sst": {
            "long_name": "Fitted sea water temperature",
            "units": "Kelvin",
            "comments": "Estimated near-surface sea water temperature from drifting buoy measurements. It is the sum of the fitted near-surface non-diurnal sea water temperature and fitted diurnal sea water temperature anomaly. Discrepancies may occur because of rounding.",
        },
        "sst1": {
            "long_name": "Fitted non-diurnal sea water temperature",
            "units": "Kelvin",
            "comments": "Estimated near-surface non-diurnal sea water temperature from drifting buoy measurements",
        },
        "sst2": {
            "long_name": "Fitted diurnal sea water temperature anomaly",
            "units": "Kelvin",
            "comments": "Estimated near-surface diurnal sea water temperature anomaly from drifting buoy measurements",
        },
        "err_sst": {
            "long_name": "Standard uncertainty of fitted sea water temperature",
            "units": "Kelvin",
            "comments": "Estimated one standard error of near-surface sea water temperature estimate from drifting buoy measurements",
        },
        "err_sst1": {
            "long_name": "Standard uncertainty of fitted non-diurnal sea water temperature",
            "units": "Kelvin",
            "comments": "Estimated one standard error of near-surface non-diurnal sea water temperature estimate from drifting buoy measurements",
        },
        "err_sst2": {
            "long_name": "Standard uncertainty of fitted diurnal sea water temperature anomaly",
            "units": "Kelvin",
            "comments": "Estimated one standard error of near-surface diurnal sea water temperature anomaly estimate from drifting buoy measurements",
        },
        "flg_sst": {
            "long_name": "Fitted sea water temperature quality flag",
            "units": "-",
            "flag_values": "0, 1, 2, 3, 4, 5",
            "flag_meanings": "no-estimate, no-uncertainty-estimate, estimate-not-in-range-uncertainty-not-in-range, estimate-not-in-range-uncertainty-in-range estimate-in-range-uncertainty-not-in-range, estimate-in-range-uncertainty-in-range",
        },
        "flg_sst1": {
            "long_name": "Fitted non-diurnal sea water temperature quality flag",
            "units": "-",
            "flag_values": "0, 1, 2, 3, 4, 5",
            "flag_meanings": "no-estimate, no-uncertainty-estimate, estimate-not-in-range-uncertainty-not-in-range, estimate-not-in-range-uncertainty-in-range estimate-in-range-uncertainty-not-in-range, estimate-in-range-uncertainty-in-range",
        },
        "flg_sst2": {
            "long_name": "Fitted diurnal sea water temperature anomaly quality flag",
            "units": "-",
            "flag_values": "0, 1, 2, 3, 4, 5",
            "flag_meanings": "no-estimate, no-uncertainty-estimate, estimate-not-in-range-uncertainty-not-in-range, estimate-not-in-range-uncertainty-in-range estimate-in-range-uncertainty-not-in-range, estimate-in-range-uncertainty-in-range",
        },
    }

    # global attributes
    attrs = {
        "title": "Global Drifter Program hourly drifting buoy collection",
        "history": f"version {GDP_VERSION}. Metadata from dirall.dat and deplog.dat",
        "Conventions": "CF-1.6",
        "time_coverage_start": "",
        "time_coverage_end": "",
        "date_created": datetime.now().isoformat(),
        "publisher_name": "GDP Drifter DAC",
        "publisher_email": "aoml.dftr@noaa.gov",
        "publisher_url": "https://www.aoml.noaa.gov/phod/gdp",
        "license": "freely available",
        "processing_level": "Level 2 QC by GDP drifter DAC",
        "metadata_link": "https://www.aoml.noaa.gov/phod/dac/dirall.html",
        "contributor_name": "NOAA Global Drifter Program",
        "contributor_role": "Data Acquisition Center",
        "institution": "NOAA Atlantic Oceanographic and Meteorological Laboratory",
        "acknowledgement": "Elipot, Shane; Sykulski, Adam; Lumpkin, Rick; Centurioni, Luca; Pazos, Mayra (2022). Hourly location, current velocity, and temperature collected from Global Drifter Program drifters world-wide. [indicate subset used]. NOAA National Centers for Environmental Information. Dataset. https://doi.org/10.25921/x46c-3620. Accessed [date]. Elipot et al. (2022): A Dataset of Hourly Sea Surface Temperature From Drifting Buoys, Scientific Data, 9, 567, https://dx.doi.org/10.1038/s41597-022-01670-2. Elipot et al. (2016): A global surface drifter dataset at hourly resolution, J. Geophys. Res.-Oceans, 121, https://dx.doi.org/10.1002/2016JC011716.",
        "summary": "Global Drifter Program hourly data",
        "doi": "10.25921/x46c-3620",
    }

    # set attributes
    for var in vars_attrs.keys():
        if var in ds.keys():
            ds[var].attrs = vars_attrs[var]
        else:
            warnings.warn(f"Variable {var} not found in upstream data; skipping.")
    ds.attrs = attrs

    # rename variables
    ds = ds.rename_vars({"longitude": "lon", "latitude": "lat"})

    # Cast float64 variables to float32 to reduce memory footprint.
    ds = gdp.cast_float64_variables_to_float32(ds)

    return ds


def to_raggedarray(
    drifter_ids: Optional[list[int]] = None,
    n_random_id: Optional[int] = None,
    url: Optional[str] = GDP_DATA_URL,
    tmp_path: Optional[str] = None,
) -> RaggedArray:
    """Download and process individual GDP hourly files and return a RaggedArray
    instance with the data.

    Parameters
    ----------
    drifter_ids : list[int], optional
        List of drifters to retrieve (Default: all)
    n_random_id : list[int], optional
        Randomly select n_random_id drifter NetCDF files
    url : str, optional
        URL from which to download the data (Default: GDP_DATA_URL).
        Alternatively, it can be GDP_DATA_URL_EXPERIMENTAL.
    tmp_path : str, optional
        Path to the directory where the individual NetCDF files are stored
        (default varies depending on operating system; /tmp/clouddrift/gdp on Linux)

    Returns
    -------
    out : RaggedArray
        A RaggedArray instance of the requested dataset

    Examples
    --------

    Invoke `to_raggedarray` without any arguments to download all drifter data
    from the 2.01 GDP feed:

    >>> from clouddrift.adapters.gdp1h import to_raggedarray
    >>> ra = to_raggedarray()

    To download a random sample of 100 drifters, for example for development
    or testing, use the `n_random_id` argument:

    >>> ra = to_raggedarray(n_random_id=100)

    To download a specific list of drifters, use the `drifter_ids` argument:

    >>> ra = to_raggedarray(drifter_ids=[44136, 54680, 83463])

    To download the experimental 2.01 GDP feed, use the `url` argument to
    specify the experimental feed URL:

    >>> from clouddrift.adapters.gdp1h import GDP_DATA_URL_EXPERIMENTAL, to_raggedarray
    >>> ra = to_raggedarray(url=GDP_DATA_URL_EXPERIMENTAL)

    Finally, `to_raggedarray` returns a `RaggedArray` instance which provides
    a convenience method to emit a `xarray.Dataset` instance:

    >>> ds = ra.to_xarray()

    To write the ragged array dataset to a NetCDF file on disk, do

    >>> ds.to_netcdf("gdp1h.nc", format="NETCDF4")

    Alternatively, to write the ragged array to a Parquet file, first create
    it as an Awkward Array:

    >>> arr = ra.to_awkward()
    >>> arr.to_parquet("gdp1h.parquet")
    """

    # adjust the tmp_path if using the experimental source
    if tmp_path is None:
        tmp_path = GDP_TMP_PATH if url == GDP_DATA_URL else GDP_TMP_PATH_EXPERIMENTAL

    ids = download(drifter_ids, n_random_id, url, tmp_path)

    if url == GDP_DATA_URL:
        filename_pattern = "drifter_hourly_{id}.nc"
    elif url == GDP_DATA_URL_EXPERIMENTAL:
        filename_pattern = "drifter_hourly_{id}.nc"
    else:
        raise ValueError(f"url must be {GDP_DATA_URL} or {GDP_DATA_URL_EXPERIMENTAL}.")

    ra = RaggedArray.from_files(
        indices=ids,
        preprocess_func=preprocess,
        name_coords=gdp.GDP_COORDS,
        name_meta=gdp.GDP_METADATA,
        name_data=GDP_DATA,
        rowsize_func=gdp.rowsize,
        filename_pattern=filename_pattern,
        tmp_path=tmp_path,
    )

    # set dynamic global attributes
    ra.attrs_global[
        "time_coverage_start"
    ] = f"{datetime(1970,1,1) + timedelta(seconds=int(np.min(ra.coords['time']))):%Y-%m-%d:%H:%M:%SZ}"
    ra.attrs_global[
        "time_coverage_end"
    ] = f"{datetime(1970,1,1) + timedelta(seconds=int(np.max(ra.coords['time']))):%Y-%m-%d:%H:%M:%SZ}"

    return ra



"""
This module provides functions to easily access ragged array datasets. If the datasets are 
not accessed via cloud storage platforms or are not found on the local filesystem,
they will be downloaded from their upstream repositories and stored for later access 
(~/.clouddrift for UNIX-based systems).
"""
from io import BufferedReader, BytesIO
from clouddrift import adapters
import os
import platform
import xarray as xr


def gdp1h(decode_times: bool = True) -> xr.Dataset:
    """Returns the latest version of the NOAA Global Drifter Program (GDP) hourly
    dataset as a ragged array Xarray dataset.

    The data is accessed from zarr archive hosted on a public AWS S3 bucket accessible at
    https://registry.opendata.aws/noaa-oar-hourly-gdp/. Original data source from NOAA NCEI
    is https://doi.org/10.25921/x46c-3620).

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        Hourly GDP dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import gdp1h
    >>> ds = gdp1h()
    >>> ds
    <xarray.Dataset>
    Dimensions:                (traj: 19396, obs: 197214787)
    Coordinates:
        id                     (traj) int64 ...
        time                   (obs) datetime64[ns] ...
    Dimensions without coordinates: traj, obs
    Data variables: (12/60)
        BuoyTypeManufacturer   (traj) |S20 ...
        BuoyTypeSensorArray    (traj) |S20 ...
        CurrentProgram         (traj) float32 ...
        DeployingCountry       (traj) |S20 ...
        DeployingShip          (traj) |S20 ...
        DeploymentComments     (traj) |S20 ...
        ...                     ...
        start_lat              (traj) float32 ...
        start_lon              (traj) float32 ...
        typebuoy               (traj) |S10 ...
        typedeath              (traj) int8 ...
        ve                     (obs) float32 ...
        vn                     (obs) float32 ...
    Attributes: (12/16)
        Conventions:       CF-1.6
        acknowledgement:   Elipot, Shane; Sykulski, Adam; Lumpkin, Rick; Centurio...
        contributor_name:  NOAA Global Drifter Program
        contributor_role:  Data Acquisition Center
        date_created:      2023-09-08T17:05:12.130123
        doi:               10.25921/x46c-3620
        ...                ...
        processing_level:  Level 2 QC by GDP drifter DAC
        publisher_email:   aoml.dftr@noaa.gov
        publisher_name:    GDP Drifter DAC
        publisher_url:     https://www.aoml.noaa.gov/phod/gdp
        summary:           Global Drifter Program hourly data
        title:             Global Drifter Program hourly drifting buoy collection

    See Also
    --------
    :func:`gdp6h`
    """
    url = "https://noaa-oar-hourly-gdp-pds.s3.amazonaws.com/latest/gdp-v2.01.zarr"
    ds = xr.open_dataset(url, engine="zarr", decode_times=decode_times)
    ds = ds.rename_vars({"ID": "id"}).assign_coords({"id": ds.ID}).drop_vars(["ids"])
    return ds


def gdp6h(decode_times: bool = True) -> xr.Dataset:
    """Returns the NOAA Global Drifter Program (GDP) 6-hourly dataset as a ragged array
    Xarray dataset.

    The data is accessed from a public HTTPS server at NOAA's Atlantic
    Oceanographic and Meteorological Laboratory (AOML) accessible at
    https://www.aoml.noaa.gov/phod/gdp/index.php. It should be noted that the data loading
    method is platform dependent. Linux and Darwin (macOS) machines lazy load the datasets leveraging the
    byte-range feature of the netCDF-c library (dataset loading engine used by xarray).
    Windows machines download the entire dataset into a memory buffer which is then passed
    to xarray.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        6-hourly GDP dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import gdp6h
    >>> ds = gdp6h()
    >>> ds
    <xarray.Dataset>
    Dimensions:                (traj: 27647, obs: 46535470)
    Coordinates:
        ids                    (obs) int64 7702204 7702204 ... 300234061198840
        time                   (obs) float64 2.879e+08 2.879e+08 ... 1.697e+09
    Dimensions without coordinates: traj, obs
    Data variables: (12/50)
        ID                     (traj) int64 7702204 7702201 ... 300234061198840
        rowsize                (traj) int32 92 1747 1943 1385 1819 ... 54 53 51 28
        WMO                    (traj) int32 0 0 0 0 ... 6203890 6203888 4101885
        expno                  (traj) int32 40 40 40 40 ... 31412 21421 21421 31412
        deploy_date            (traj) float32 2.878e+08 2.878e+08 ... 1.696e+09 nan
        deploy_lat             (traj) float32 -7.798 -4.9 -3.18 ... 9.9 11.9 nan
        ...                     ...
        vn                     (obs) float32 nan 0.1056 0.04974 ... 0.7384 nan
        temp                   (obs) float32 28.35 28.3 nan ... 29.08 28.97 28.92
        err_lat                (obs) float32 0.009737 0.007097 ... 0.001659 0.001687
        err_lon                (obs) float32 0.00614 0.004583 ... 0.002471 0.002545
        err_temp               (obs) float32 0.08666 0.08757 ... 0.03665 0.03665
        drogue_status          (obs) bool False False False False ... True True True
    Attributes: (12/18)
        title:                Global Drifter Program drifting buoy collection
        history:              version September 2023. Metadata from dirall.dat an...
        Conventions:          CF-1.6
        time_coverage_start:  1979-02-15:00:00:00Z
        time_coverage_end:    2023-10-18:18:00:00Z
        date_created:         2023-12-22T17:50:22.242943
        ...                   ...
        contributor_name:     NOAA Global Drifter Program
        contributor_role:     Data Acquisition Center
        institution:          NOAA Atlantic Oceanographic and Meteorological Labo...
        acknowledgement:      Lumpkin, Rick; Centurioni, Luca (2019). NOAA Global...
        summary:              Global Drifter Program six-hourly data
        doi:                  10.25921/7ntx-z961

    See Also
    --------
    :func:`gdp1h`
    """
    url = "https://www.aoml.noaa.gov/ftp/pub/phod/buoydata/gdp6h_ragged_may23.nc#mode=bytes"

    if platform.system() == "Windows":
        buffer = BytesIO()
        adapters.utils.download_with_progress([(f"{url}#mode=bytes", buffer)])
        reader = BufferedReader(buffer)
        ds = xr.open_dataset(reader, decode_times=decode_times)
    else:
        ds = xr.open_dataset(f"{url}", decode_times=decode_times)

    ds = ds.rename_vars({"ID": "id"}).assign_coords({"id": ds.ID}).drop_vars(["ids"])
    return ds


def glad(decode_times: bool = True) -> xr.Dataset:
    """Returns the Grand LAgrangian Deployment (GLAD) dataset as a ragged array
    Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access.

    The upstream data is available at https://doi.org/10.7266/N7VD6WC8.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        GLAD dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import glad
    >>> ds = glad()
    >>> ds
    <xarray.Dataset>
    Dimensions:         (obs: 1602883, traj: 297)
    Coordinates:
      time            (obs) datetime64[ns] ...
      id              (traj) object ...
    Data variables:
      latitude        (obs) float32 ...
      longitude       (obs) float32 ...
      position_error  (obs) float32 ...
      u               (obs) float32 ...
      v               (obs) float32 ...
      velocity_error  (obs) float32 ...
      rowsize         (traj) int64 ...
    Attributes:
      title:        GLAD experiment CODE-style drifter trajectories (low-pass f...
      institution:  Consortium for Advanced Research on Transport of Hydrocarbo...
      source:       CODE-style drifters
      history:      Downloaded from https://data.gulfresearchinitiative.org/dat...
      references:   Özgökmen, Tamay. 2013. GLAD experiment CODE-style drifter t...

    Reference
    ---------
    Özgökmen, Tamay. 2013. GLAD experiment CODE-style drifter trajectories (low-pass filtered, 15 minute interval records), northern Gulf of Mexico near DeSoto Canyon, July-October 2012. Distributed by: Gulf of Mexico Research Initiative Information and Data Cooperative (GRIIDC), Harte Research Institute, Texas A&M University–Corpus Christi. doi:10.7266/N7VD6WC8
    """
    clouddrift_path = (
        os.path.expanduser("~/.clouddrift")
        if not os.getenv("CLOUDDRIFT_PATH")
        else os.getenv("CLOUDDRIFT_PATH")
    )
    glad_path = f"{clouddrift_path}/data/glad.nc"
    if not os.path.exists(glad_path):
        print(f"{glad_path} not found; download from upstream repository.")
        ds = adapters.glad.to_xarray()
        os.makedirs(os.path.dirname(glad_path), exist_ok=True)
        ds.to_netcdf(glad_path)
    else:
        ds = xr.open_dataset(glad_path, decode_times=decode_times)
    return ds


def mosaic(decode_times: bool = True) -> xr.Dataset:
    """Returns the MOSAiC sea-ice drift dataset as a ragged array Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access.

    The upstream data is available at https://arcticdata.io/catalog/view/doi:10.18739/A2KP7TS83.

    Reference
    ---------
    Angela Bliss, Jennifer Hutchings, Philip Anderson, Philipp Anhaus,
    Hans Jakob Belter, Jørgen Berge, Vladimir Bessonov, Bin Cheng, Sylvia Cole,
    Dave Costa, Finlo Cottier, Christopher J Cox, Pedro R De La Torre, Dmitry V Divine,
    Gilbert Emzivat, Ying-Chih Fang, Steven Fons, Michael Gallagher, Maxime Geoffrey,
    Mats A Granskog, ... Guangyu Zuo. (2022). Sea ice drift tracks from the Distributed
    Network of autonomous buoys deployed during the Multidisciplinary drifting Observatory
    for the Study of Arctic Climate (MOSAiC) expedition 2019 - 2021. Arctic Data Center.
    doi:10.18739/A2KP7TS83.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        MOSAiC sea-ice drift dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import mosaic
    >>> ds = mosaic()
    >>> ds
    <xarray.Dataset>
    Dimensions:                     (obs: 1926226, traj: 216)
    Coordinates:
        time                        (obs) datetime64[ns] ...
        id                          (traj) object ...
    Dimensions without coordinates: obs, traj
    Data variables: (12/19)
        latitude                    (obs) float64 ...
        longitude                   (obs) float64 ...
        Deployment Leg              (traj) int64 ...
        DN Station ID               (traj) object ...
        IMEI                        (traj) object ...
        Deployment Date             (traj) datetime64[ns] ...
        ...                          ...
        Buoy Type                   (traj) object ...
        Manufacturer                (traj) object ...
        Model                       (traj) object ...
        PI                          (traj) object ...
        Data Authors                (traj) object ...
        rowsize                     (traj) int64 ...
    """
    clouddrift_path = (
        os.path.expanduser("~/.clouddrift")
        if not os.getenv("CLOUDDRIFT_PATH")
        else os.getenv("CLOUDDRIFT_PATH")
    )
    mosaic_path = f"{clouddrift_path}/data/mosaic.nc"
    if not os.path.exists(mosaic_path):
        print(f"{mosaic_path} not found; download from upstream repository.")
        ds = adapters.mosaic.to_xarray()
        os.makedirs(os.path.dirname(mosaic_path), exist_ok=True)
        ds.to_netcdf(mosaic_path)
    else:
        ds = xr.open_dataset(mosaic_path, decode_times=decode_times)
    return ds


def spotters(decode_times: bool = True) -> xr.Dataset:
    """Returns the Sofar Ocean Spotter drifters ragged array dataset as an Xarray dataset.

    The data is accessed from a zarr archive hosted on a public AWS S3 bucket accessible
    at https://sofar-spotter-archive.s3.amazonaws.com/spotter_data_bulk_zarr.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        Sofar ocean floats dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import spotters
    >>> ds = spotters()
    >>> ds
    <xarray.Dataset>
    Dimensions:                (index: 6390651, trajectory: 871)
    Coordinates:
        time                   (index) datetime64[ns] ...
      * trajectory             (trajectory) object 'SPOT-010001' ... 'SPOT-1975'
    Dimensions without coordinates: index
    Data variables:
        latitude               (index) float64 ...
        longitude              (index) float64 ...
        meanDirection          (index) float64 ...
        meanDirectionalSpread  (index) float64 ...
        meanPeriod             (index) float64 ...
        peakDirection          (index) float64 ...
        peakDirectionalSpread  (index) float64 ...
        peakPeriod             (index) float64 ...
        rowsize                (trajectory) int64 ...
        significantWaveHeight  (index) float64 ...
    Attributes:
        author:         Isabel A. Houghton
        creation_date:  2023-10-18 00:43:55.333537
        email:          isabel.houghton@sofarocean.com
        institution:    Sofar Ocean
        references:     https://content.sofarocean.com/hubfs/Spotter%20product%20...
        source:         Spotter wave buoy
        title:          Sofar Spotter Data Archive - Bulk Wave Parameters
    """
    url = "https://sofar-spotter-archive.s3.amazonaws.com/spotter_data_bulk_zarr"
    return xr.open_dataset(url, engine="zarr", decode_times=decode_times)


def subsurface_floats(decode_times: bool = True) -> xr.Dataset:
    """Returns the subsurface floats dataset as a ragged array Xarray dataset.

    The data is accessed from a public HTTPS server at NOAA's Atlantic
    Oceanographic and Meteorological Laboratory (AOML) accessible at
    https://www.aoml.noaa.gov/phod/gdp/index.php.

    The upstream data is available at
    https://www.aoml.noaa.gov/phod/float_traj/files/allFloats_12122017.mat.

    This dataset of subsurface float observations was compiled by the WOCE Subsurface
    Float Data Assembly Center (WFDAC) in Woods Hole maintained by Andree Ramsey and
    Heather Furey and copied to NOAA/AOML in October 2014 (version 1) and in December
    2017 (version 2). Subsequent updates will be included as additional appropriate
    float data, quality controlled by the appropriate principal investigators, is
    submitted for inclusion.

    Note that these observations are collected by ALACE/RAFOS/Eurofloat-style
    acoustically-tracked, neutrally-buoyant subsurface floats which collect data while
    drifting beneath the ocean surface. These data are the result of the effort and
    resources of many individuals and institutions. You are encouraged to acknowledge
    the work of the data originators and Data Centers in publications arising from use
    of these data.

    The float data were originally divided by project at the WFDAC. Here they have been
    compiled in a single Matlab data set. See here for more information on the variables
    contained in these files.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        Subsurface floats dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import subsurface_floats
    >>> ds = subsurface_floats()
    >>> ds
    <xarray.Dataset>
    Dimensions:   (traj: 2193, obs: 1402840)
    Coordinates:
        id        (traj) uint16 ...
        time      (obs) datetime64[ns] ...
    Dimensions without coordinates: traj, obs
    Data variables: (12/13)
        expList   (traj) object ...
        expName   (traj) object ...
        expOrg    (traj) object ...
        expPI     (traj) object ...
        indexExp  (traj) uint8 ...
        fltType   (traj) object ...
        ...        ...
        lon       (obs) float64 ...
        lat       (obs) float64 ...
        pres      (obs) float64 ...
        temp      (obs) float64 ...
        ve        (obs) float64 ...
        vn        (obs) float64 ...
    Attributes:
        title:            Subsurface float trajectories dataset
        history:          December 2017 (version 2)
        date_created:     2023-11-14T22:30:38.831656
        publisher_name:   WOCE Subsurface Float Data Assembly Center and NOAA AOML
        publisher_url:    https://www.aoml.noaa.gov/phod/float_traj/data.php
        license:          freely available
        acknowledgement:  Maintained by Andree Ramsey and Heather Furey from the ...

    References
    ----------
    WOCE Subsurface Float Data Assembly Center (WFDAC) https://www.aoml.noaa.gov/phod/float_traj/index.php
    """

    clouddrift_path = (
        os.path.expanduser("~/.clouddrift")
        if not os.getenv("CLOUDDRIFT_PATH")
        else os.getenv("CLOUDDRIFT_PATH")
    )

    local_file = f"{clouddrift_path}/data/subsurface_floats.nc"
    if not os.path.exists(local_file):
        print(f"{local_file} not found; download from upstream repository.")
        ds = adapters.subsurface_floats.to_xarray()
    else:
        ds = xr.open_dataset(local_file, decode_times=decode_times)
    return ds


def yomaha(decode_times: bool = True) -> xr.Dataset:
    """Returns the YoMaHa dataset as a ragged array Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access. The upstream
    data is available at http://apdrc.soest.hawaii.edu/projects/yomaha/.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        YoMaHa'07 dataset as a ragged array

    Examples
    --------

    >>> from clouddrift.datasets import yomaha
    >>> ds = yomaha()
    >>> ds
    <xarray.Dataset>
    Dimensions:     (obs: 1926743, traj: 12196)
    Coordinates:
        time_d      (obs) datetime64[ns] ...
        time_s      (obs) datetime64[ns] ...
        time_lp     (obs) datetime64[ns] ...
        time_lc     (obs) datetime64[ns] ...
        id          (traj) int64 ...
    Dimensions without coordinates: obs, traj
    Data variables: (12/27)
        lon_d       (obs) float64 ...
        lat_d       (obs) float64 ...
        pres_d      (obs) float32 ...
        ve_d        (obs) float32 ...
        vn_d        (obs) float32 ...
        err_ve_d    (obs) float32 ...
        ...          ...
        cycle       (obs) int64 ...
        time_inv    (obs) int64 ...
        rowsize     (traj) int64 ...
        wmo_id      (traj) int64 ...
        dac_id      (traj) int64 ...
        float_type  (traj) int64 ...
    Attributes:
        title:           YoMaHa'07: Velocity data assessed from trajectories of A...
        history:         Dataset updated on Tue Jun 28 03:14:34 HST 2022
        date_created:    2023-12-08T00:52:08.478075
        publisher_name:  Asia-Pacific Data Research Center
        publisher_url:   http://apdrc.soest.hawaii.edu/index.php
        license:         Creative Commons Attribution 4.0 International License..

    Reference
    ---------
    Lebedev, K. V., Yoshinari, H., Maximenko, N. A., & Hacker, P. W. (2007). Velocity data
    assessed  from trajectories of Argo floats at parking level and at the sea
    surface. IPRC Technical Note, 4(2), 1-16.
    """
    clouddrift_path = (
        os.path.expanduser("~/.clouddrift")
        if not os.getenv("CLOUDDRIFT_PATH")
        else os.getenv("CLOUDDRIFT_PATH")
    )
    local_file = f"{clouddrift_path}/data/yomaha.nc"
    if not os.path.exists(local_file):
        print(f"{local_file} not found; download from upstream repository.")
        ds = adapters.yomaha.to_xarray()
        os.makedirs(os.path.dirname(local_file), exist_ok=True)
        ds.to_netcdf(local_file)
    else:
        ds = xr.open_dataset(local_file, decode_times=decode_times)
    return ds


def andro(decode_times: bool = True) -> xr.Dataset:
    """Returns the ANDRO as a ragged array Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access. The upstream
    data is available at https://www.seanoe.org/data/00360/47077/.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        ANDRO dataset as a ragged array
    Examples
    --------
    >>> from clouddrift.datasets import andro
    >>> ds = andro()
    >>> ds
    <xarray.Dataset>
    Dimensions:     (obs: 1360753, traj: 9996)
    Coordinates:
        time_d      (obs) datetime64[ns] ...
        time_s      (obs) datetime64[ns] ...
        time_lp     (obs) datetime64[ns] ...
        time_lc     (obs) datetime64[ns] ...
        id          (traj) int64 ...
    Dimensions without coordinates: obs, traj
    Data variables: (12/33)
        lon_d       (obs) float64 ...
        lat_d       (obs) float64 ...
        pres_d      (obs) float32 ...
        temp_d      (obs) float32 ...
        sal_d       (obs) float32 ...
        ve_d        (obs) float32 ...
        ...          ...
        lon_lc      (obs) float64 ...
        lat_lc      (obs) float64 ...
        surf_fix    (obs) int64 ...
        cycle       (obs) int64 ...
        profile_id  (obs) float32 ...
        rowsize     (traj) int64 ...
    Attributes:
        title:           ANDRO: An Argo-based deep displacement dataset
        history:         2022-03-04
        date_created:    2023-12-08T00:52:00.937120
        publisher_name:  SEANOE (SEA scieNtific Open data Edition)
        publisher_url:   https://www.seanoe.org/data/00360/47077/
        license:         freely available

    Reference
    ---------
    Ollitrault Michel, Rannou Philippe, Brion Emilie, Cabanes Cecile, Piron Anne, Reverdin Gilles,
    Kolodziejczyk Nicolas (2022). ANDRO: An Argo-based deep displacement dataset.
    SEANOE. https://doi.org/10.17882/47077
    """
    clouddrift_path = (
        os.path.expanduser("~/.clouddrift")
        if not os.getenv("CLOUDDRIFT_PATH")
        else os.getenv("CLOUDDRIFT_PATH")
    )
    local_file = f"{clouddrift_path}/data/andro.nc"
    if not os.path.exists(local_file):
        print(f"{local_file} not found; download from upstream repository.")
        ds = adapters.andro.to_xarray()
        os.makedirs(os.path.dirname(local_file), exist_ok=True)
        ds.to_netcdf(local_file)
    else:
        ds = xr.open_dataset(local_file, decode_times=decode_times)
    return ds



"""
This module defines functions used to adapt the ANDRO: An Argo-based 
deep displacement dataset as a ragged-arrays dataset. 

The dataset is hosted at https://www.seanoe.org/data/00360/47077/ and the user manual
is available at https://archimer.ifremer.fr/doc/00360/47126/.

Example
-------
>>> from clouddrift.adapters import andro
>>> ds = andro.to_xarray()

Reference
---------
Ollitrault Michel, Rannou Philippe, Brion Emilie, Cabanes Cecile, Piron Anne, Reverdin Gilles, 
Kolodziejczyk Nicolas (2022). ANDRO: An Argo-based deep displacement dataset. 
SEANOE. https://doi.org/10.17882/47077
"""

from clouddrift.adapters.utils import download_with_progress
from datetime import datetime
import numpy as np
import os
import pandas as pd
import tempfile
import xarray as xr
import warnings

# order of the URLs is important
ANDRO_URL = "https://www.seanoe.org/data/00360/47077/data/91950.dat"
ANDRO_TMP_PATH = os.path.join(tempfile.gettempdir(), "clouddrift", "andro")
ANDRO_VERSION = "2022-03-04"


def to_xarray(tmp_path: str = None):
    if tmp_path is None:
        tmp_path = ANDRO_TMP_PATH
        os.makedirs(tmp_path, exist_ok=True)

    # get or update dataset
    local_file = f"{tmp_path}/{ANDRO_URL.split('/')[-1]}"
    download_with_progress([(ANDRO_URL, local_file)])

    # parse with panda
    col_names = [
        # depth
        "lon_d",
        "lat_d",
        "pres_d",
        "temp_d",
        "sal_d",
        "time_d",
        "ve_d",
        "vn_d",
        "err_ve_d",
        "err_vn_d",
        # first surface velocity
        "lon_s",
        "lat_s",
        "time_s",
        "ve_s",
        "vn_s",
        "err_ve_s",
        "err_vn_s",
        # last surface velocity
        "lon_ls",
        "lat_ls",
        "time_ls",
        "ve_ls",
        "vn_ls",
        "err_ve_ls",
        "err_vn_ls",
        # last fix previous cycle
        "lon_lp",
        "lat_lp",
        "time_lp",
        # first fix current cycle
        "lon_fc",
        "lat_fc",
        "time_fc",
        # last fix current cycle
        "lon_lc",
        "lat_lc",
        "time_lc",
        "surf_fix",
        "id",
        "cycle",
        "profile_id",
    ]

    na_col = [
        -999.9999,
        -99.9999,
        -999.9,
        -99.999,
        -99.999,
        -9999.999,
        -999.9,
        -999.9,
        -999.9,
        -999.9,
        -999.9999,
        -99.999,
        -9999.999,
        -999.99,
        -999.99,
        -999.99,
        -999.99,
        -999.9999,
        -99.9999,
        -9999.999,
        -999.99,
        -999.99,
        -999.99,
        -999.99,
        -999.9999,
        -99.9999,
        -9999.999,
        -999.9999,
        -99.9999,
        -9999.999,
        -999.9999,
        -99.9999,
        -9999.999,
        np.nan,
        np.nan,
        np.nan,
        -99,
    ]

    # open with pandas
    df = pd.read_csv(
        local_file, names=col_names, sep="\s+", header=None, na_values=na_col
    )

    # convert to an Xarray Dataset
    ds = xr.Dataset.from_dataframe(df)

    unique_id, rowsize = np.unique(ds["id"], return_counts=True)

    ds = (
        ds.rename_dims({"index": "obs"})
        .assign({"id": ("traj", unique_id)})
        .assign({"rowsize": ("traj", rowsize)})
        .set_coords(["id", "time_d", "time_s", "time_lp", "time_lc", "time_lp"])
        .drop_vars(["index"])
    )

    # Cast double floats to singles
    double_vars = [
        "lat_d",
        "lon_d",
        "lat_s",
        "lon_s",
        "lat_ls",
        "lon_ls",
        "lat_lp",
        "lon_lp",
        "lat_fc",
        "lon_fc",
        "lat_lc",
        "lon_lc",
    ]
    for var in [v for v in ds.variables if v not in double_vars]:
        if ds[var].dtype == "float64":
            ds[var] = ds[var].astype("float32")

    # define attributes
    vars_attrs = {
        "lon_d": {
            "long_name": "Longitude of the location where the deep velocity is calculated",
            "units": "degrees_east",
        },
        "lat_d": {
            "long_name": "Latitude of the location where the deep velocity is calculated",
            "units": "degrees_north",
        },
        "pres_d": {
            "long_name": "Reference parking pressure for this cycle",
            "units": "dbar",
        },
        "temp_d": {
            "long_name": "Parking temperature (°C) for this cycle",
            "units": "degree_C",
        },
        "sal_d": {
            "long_name": "Parking salinity for this cycle",
            "units": "psu",
        },
        "time_d": {
            "long_name": "Julian time (days) when deep velocity is estimated",
            "units": "days since 2000-01-01 00:00",
        },
        "ve_d": {
            "long_name": "Eastward component of the deep velocity",
            "units": "cm s-1",
        },
        "vn_d": {
            "long_name": "Northward component of the deep velocity",
            "units": "cm s-1",
        },
        "err_ve_d": {
            "long_name": "Error on the eastward component of the deep velocity",
            "units": "cm s-1",
        },
        "err_vn_d": {
            "long_name": "Error on the northward component of the deep velocity",
            "units": "cm s-1",
        },
        "lon_s": {
            "long_name": "Longitude of the location where the first surface velocity is calculated (over the first 6 h at surface)",
            "units": "degrees_east",
        },
        "lat_s": {
            "long_name": "Latitude of the location where the first surface velocity is calculated",
            "units": "degrees_north",
        },
        "time_s": {
            "long_name": "Julian time (days) when the first surface velocity is calculated",
            "units": "days since 2000-01-01 00:00",
        },
        "ve_s": {
            "long_name": "Eastward component of first surface velocity",
            "units": "cm s-1",
        },
        "vn_s": {
            "long_name": "Northward component of first surface velocity",
            "units": "cm s-1",
        },
        "err_ve_s": {
            "long_name": "Error on the eastward component of the first surface velocity",
            "units": "cm s-1",
        },
        "err_vn_s": {
            "long_name": "Error on the northward component of the first surface velocity",
            "units": "cm s-1",
        },
        "lon_ls": {
            "long_name": "Longitude of the location where the last surface velocity is calculated (over the last 6 h at surface)",
            "units": "degrees_east",
        },
        "lat_ls": {
            "long_name": "Latitude of the location where the last surface velocity is calculated",
            "units": "degrees_north",
        },
        "time_ls": {
            "long_name": "Julian time (days) when the last surface velocity is calculated",
            "units": "days since 2000-01-01 00:00",
        },
        "ve_ls": {
            "long_name": "Eastward component of last surface velocity (cm s-1)",
            "units": "cm s-1",
        },
        "vn_ls": {
            "long_name": "Northward component of last surface velocity (cm s-1)",
            "units": "cm s-1",
        },
        "err_ve_ls": {
            "long_name": "Error on the eastward component of the last surface velocity",
            "units": "cm s-1",
        },
        "err_vn_ls": {
            "long_name": "Error on the northward component of the last surface velocity",
            "units": "cm s-1",
        },
        "lon_lp": {
            "long_name": "Longitude of the last fix at the sea surface during the previous cycle",
            "units": "degrees_east",
        },
        "lat_lp": {
            "long_name": "Latitude of the last fix at the sea surface during the previous cycle",
            "units": "degrees_north",
        },
        "time_lp": {
            "long_name": "Julian time of the last fix at the sea surface during the previous cycle",
            "units": "days since 2000-01-01 00:00",
        },
        "lon_fc": {
            "long_name": "Longitude of the first fix at the sea surface during the current cycle",
            "units": "degrees_east",
        },
        "lat_fc": {
            "long_name": "Latitude of the first fix at the sea surface during the current cycle",
            "units": "degrees_north",
        },
        "time_fc": {
            "long_name": "Julian time of the first fix at the sea surface during the current cycle",
            "units": "days since 2000-01-01 00:00",
        },
        "lon_lc": {
            "long_name": "Longitude of the last fix at the sea surface during the current cycle",
            "units": "degrees_east",
        },
        "lat_lc": {
            "long_name": "Latitude of the last fix at the sea surface during the current cycle",
            "units": "degrees_north",
        },
        "time_lc": {
            "long_name": "Julian time of the last fix at the sea surface during the current cycle",
            "units": "days since 2000-01-01 00:00",
        },
        "surf_fix": {
            "long_name": "Number of surface fixes during the current cycle",
            "units": "-",
        },
        "id": {
            "long_name": "Float WMO number",
            "units": "-",
        },
        "cycle": {
            "long_name": "Cycle number",
            "units": "-",
        },
        "profile_id": {
            "long_name": "Profile number as given in the NetCDF prof file",
            "units": "-",
        },
    }

    # global attributes
    attrs = {
        "title": "ANDRO: An Argo-based deep displacement dataset",
        "history": f"Dataset updated on {ANDRO_VERSION}",
        "date_created": datetime.now().isoformat(),
        "publisher_name": "SEANOE (SEA scieNtific Open data Edition)",
        "publisher_url": "https://www.seanoe.org/data/00360/47077/",
        "license": "Creative Commons Attribution 4.0 International License (http://creativecommons.org/licenses/by/4.0/)",
    }

    # set attributes
    for var in vars_attrs.keys():
        if var in ds.keys():
            ds[var].attrs = vars_attrs[var]
        else:
            warnings.warn(f"Variable {var} not found in upstream data; skipping.")
    ds.attrs = attrs

    return ds



"""
This module provides adapters to custom datasets.
Each adapter module provides convenience functions and metadata to convert a
custom dataset to a `clouddrift.RaggedArray` instance.
Currently, clouddrift provides adapter modules for the hourly Global
Drifter Program (GDP) dataset, the 6-hourly GDP dataset, 15-minute Grand LAgrangian
Deployment (GLAD) dataset, and the MOSAiC sea-ice drift dataset. More adapters will be added
in the future.
"""

import clouddrift.adapters.andro as andro
import clouddrift.adapters.gdp1h as gdp1h
import clouddrift.adapters.gdp6h as gdp6h
import clouddrift.adapters.glad as glad
import clouddrift.adapters.mosaic as mosaic
import clouddrift.adapters.subsurface_floats as subsurface_floats
import clouddrift.adapters.yomaha as yomaha
import clouddrift.adapters.utils as utils


__all__ = [
    "andro",
    "gdp1h",
    "gdp6h",
    "glad",
    "mosaic",
    "subsurface_floats",
    "yomaha",
    "utils",
]



"""
This module provides functions for spherical geometry calculations.
"""

import numpy as np
from typing import Optional, Tuple, Union
import xarray as xr
import warnings

EARTH_RADIUS_METERS = 6.3781e6
EARTH_DAY_SECONDS = 86164.091
EARTH_ROTATION_RATE = 2 * np.pi / EARTH_DAY_SECONDS


def cumulative_distance(
    longitude: Union[list, np.ndarray, xr.DataArray],
    latitude: Union[list, np.ndarray, xr.DataArray],
) -> np.ndarray:
    """Return the cumulative great circle distance in meters along a sequence of geographical locations.

    Parameters
    ----------
    latitude : array-like
        Latitude sequence, in degrees.
    longitude : array-like
        Longitude sequence, in degrees.

    Returns
    -------
    out : np.ndarray
        Cumulative distance.

    See Also
    --------
    :func:`distance`

    Examples
    --------
    Calculate the cumulative distance in meters along a path of three points:

    >>> cumulative_distance(np.array([0, 1, 2]), np.array([0, 1, 2]))
    array([     0.        , 157424.62387233, 314825.27182116])
    """
    return np.cumsum(
        np.concatenate(
            (
                [0],
                distance(latitude[0:-1], longitude[0:-1], latitude[1:], longitude[1:]),
            )
        )
    )


def distance(
    lon1: Union[float, list, np.ndarray, xr.DataArray],
    lat1: Union[float, list, np.ndarray, xr.DataArray],
    lon2: Union[float, list, np.ndarray, xr.DataArray],
    lat2: Union[float, list, np.ndarray, xr.DataArray],
) -> Union[float, np.ndarray]:
    """Return elementwise great circle distance in meters between one or more
    points from arrays of their latitudes and longitudes, using the Haversine
    formula.

    d = 2⋅r⋅asin √[sin²(Δφ/2) + cos φ1 ⋅ cos φ2 ⋅ sin²(Δλ/2)]

    where (φ, λ) is (lat, lon) in radians and r is the radius of the sphere in
    meters.

    Parameters
    ----------
    lon1 : np.ndarray
        Longitudes of the first set of points, in degrees
    lat1 : np.ndarray
        Latitudes of the first set of points, in degrees
    lon2 : np.ndarray
        Longitudes of the second set of points, in degrees
    lat2 : np.ndarray
        Latitudes of the second set of points, in degrees

    Returns
    -------
    out : np.ndarray
        Great circle distance

    Examples
    --------
    Calculate the distance of one degree longitude on the equator:

    >>> distance(0, 0, 0, 1)
    111318.84502145034

    Calculate the distance of one degree longitude at 45-degrees North latitude:

    >>> distance(0, 45, 1, 45)
    78713.81064540472

    You can also pass array-like inputs to calculate an array of distances:

    >>> distance([0, 0], [0, 45], [0, 1], [1, 45])
    array([111318.84502145,  78713.8106454 ])
    """

    # Input coordinates are in degrees; convert to radians.
    # If any of the input arrays are xr.DataArray, extract the values first
    # because Xarray enforces alignment between coordinates.
    if type(lat1) is xr.DataArray:
        lat1_rad = np.deg2rad(lat1.values)
    else:
        lat1_rad = np.deg2rad(lat1)
    if type(lon1) is xr.DataArray:
        lon1_rad = np.deg2rad(lon1.values)
    else:
        lon1_rad = np.deg2rad(lon1)
    if type(lat2) is xr.DataArray:
        lat2_rad = np.deg2rad(lat2.values)
    else:
        lat2_rad = np.deg2rad(lat2)
    if type(lon2) is xr.DataArray:
        lon2_rad = np.deg2rad(lon2.values)
    else:
        lon2_rad = np.deg2rad(lon2)

    dlat = lat2_rad - lat1_rad
    dlon = lon2_rad - lon1_rad

    h = (
        np.sin(0.5 * dlat) ** 2
        + np.cos(lat1_rad) * np.cos(lat2_rad) * np.sin(0.5 * dlon) ** 2
    )

    return 2 * np.arcsin(np.sqrt(h)) * EARTH_RADIUS_METERS


def bearing(
    lon1: Union[float, list, np.ndarray, xr.DataArray],
    lat1: Union[float, list, np.ndarray, xr.DataArray],
    lon2: Union[float, list, np.ndarray, xr.DataArray],
    lat2: Union[float, list, np.ndarray, xr.DataArray],
) -> Union[float, np.ndarray]:
    """Return elementwise initial (forward) bearing in radians from arrays of
    latitude and longitude in degrees, based on the spherical law of cosines.

    The formula is:

    θ = atan2(cos φ1 ⋅ sin φ2 - sin φ1 ⋅ cos φ2 ⋅ cos Δλ, sin Δλ ⋅ cos φ2)

    where (φ, λ) is (lat, lon) and θ is bearing, all in radians.
    Bearing is defined as zero toward East and positive counterclockwise.

    Parameters
    ----------
    lon1 : float or array-like
        Longitudes of the first set of points, in degrees
    lat1 : float or array-like
        Latitudes of the first set of points, in degrees
    lon2 : float or array-like
        Longitudes of the second set of points, in degrees
    lat2 : float or array-like
        Latitudes of the second set of points, in degrees

    Returns
    -------
    theta : float or np.ndarray
        Bearing angles in radians

    Examples
    --------
    Calculate the bearing of one degree longitude on the equator:

    >>> bearing(0, 0, 1, 0)
    0.0

    Calculate the bearing of 10 degrees longitude at 45-degrees North latitude:

    >>> bearing(0, 45, 10, 45)
    0.06178508761798218
    """
    # Input coordinates are in degrees; convert to radians.
    # If any of the input arrays are xr.DataArray, extract the values first
    # because Xarray enforces alignment between coordinates.
    if type(lat1) is xr.DataArray:
        lat1_rad = np.deg2rad(lat1.values)
    else:
        lat1_rad = np.deg2rad(lat1)
    if type(lon1) is xr.DataArray:
        lon1_rad = np.deg2rad(lon1.values)
    else:
        lon1_rad = np.deg2rad(lon1)
    if type(lat2) is xr.DataArray:
        lat2_rad = np.deg2rad(lat2.values)
    else:
        lat2_rad = np.deg2rad(lat2)
    if type(lon2) is xr.DataArray:
        lon2_rad = np.deg2rad(lon2.values)
    else:
        lon2_rad = np.deg2rad(lon2)

    dlon = lon2_rad - lon1_rad

    theta = np.arctan2(
        np.cos(lat1_rad) * np.sin(lat2_rad)
        - np.sin(lat1_rad) * np.cos(lat2_rad) * np.cos(dlon),
        np.sin(dlon) * np.cos(lat2_rad),
    )

    return theta


def position_from_distance_and_bearing(
    lon: float, lat: float, distance: float, bearing: float
) -> Tuple[float, float]:
    """Return elementwise new position in degrees from arrays of latitude and
    longitude in degrees, distance in meters, and bearing in radians, based on
    the spherical law of cosines.

    The formula is:

    φ2 = asin( sin φ1 ⋅ cos δ + cos φ1 ⋅ sin δ ⋅ cos θ )
    λ2 = λ1 + atan2( sin θ ⋅ sin δ ⋅ cos φ1, cos δ − sin φ1 ⋅ sin φ2 )

    where (φ, λ) is (lat, lon) and θ is bearing, all in radians.
    Bearing is defined as zero toward East and positive counterclockwise.

    Parameters
    ----------
    lon : float
        Longitude of the first set of points, in degrees
    lat : float
        Latitude of the first set of points, in degrees
    distance : array_like
        Distance in meters
    bearing : array_like
        Bearing angles in radians

    Returns
    -------
    lon2 : array_like
        Latitudes of the second set of points, in degrees, in the range [-90, 90]
    lat2 : array_like
        Longitudes of the second set of points, in degrees, in the range [-180, 180]

    Examples
    --------
    Calculate the position of one degree longitude distance on the equator:

    >>> position_from_distance_and_bearing(0, 0, 111318.84502145034, 0)
    (1.0, 0.0)

    Calculate the position of one degree latitude distance from 45 degrees North latitude:

    >>> position_from_distance_and_bearing(0, 45, 111318.84502145034, np.pi / 2)
    (8.81429402840006e-17, 45.99999999999999)
    """
    lat_rad = np.deg2rad(lat)
    lon_rad = np.deg2rad(lon)

    distance_rad = distance / EARTH_RADIUS_METERS

    lat2_rad = np.arcsin(
        np.sin(lat_rad) * np.cos(distance_rad)
        + np.cos(lat_rad) * np.sin(distance_rad) * np.sin(bearing)
    )
    lon2_rad = lon_rad + np.arctan2(
        np.cos(bearing) * np.sin(distance_rad) * np.cos(lat_rad),
        np.cos(distance_rad) - np.sin(lat_rad) * np.sin(lat2_rad),
    )

    return np.rad2deg(lon2_rad), np.rad2deg(lat2_rad)


def recast_lon(lon: np.ndarray, lon0: Optional[float] = -180) -> np.ndarray:
    """Recast (convert) longitude values to a selected range of 360 degrees
    starting from ``lon0``.

    Parameters
    ----------
    lon : np.ndarray or float
        An N-d array of longitudes in degrees
    lon0 : float, optional
        Starting longitude of the recasted range (default -180).

    Returns
    -------
    np.ndarray or float
        Converted longitudes in the range `[lon0, lon0+360[`

    Examples
    --------
    By default, ``recast_lon`` converts longitude values to the range
    `[-180, 180[`:

    >>> recast_lon(200)
    -160

    >>> recast_lon(180)
    -180

    The range of the output longitude is controlled by ``lon0``.
    For example, with ``lon0 = 0``, the longitude values are converted to the
    range `[0, 360[`.

    >>> recast_lon(200, -180)
    -160

    With ``lon0 = 20``, longitude values are converted to range `[20, 380]`,
    which can be useful to avoid cutting the major ocean basins.

    >>> recast_lon(10, 20)
    370

    See Also
    --------
    :func:`recast_lon360`, :func:`recast_lon180`
    """
    return np.mod(lon - lon0, 360) + lon0


def recast_lon360(lon: np.ndarray) -> np.ndarray:
    """Recast (convert) longitude values to the range `[0, 360[`.
    This is a convenience wrapper around :func:`recast_lon` with ``lon0 = 0``.

    Parameters
    ----------
    lon : np.ndarray
        An N-d array of longitudes in degrees

    Returns
    -------
    np.ndarray
        Converted longitudes in the range `[0, 360[`

    Examples
    --------
    >>> recast_lon360(200)
    200

    >>> recast_lon360(-200)
    160

    See Also
    --------
    :func:`recast_lon`, :func:`recast_lon180`
    """
    return recast_lon(lon, 0)


def recast_lon180(lon: np.ndarray) -> np.ndarray:
    """Recast (convert) longitude values to the range `[-180, 180[`.
    This is a convenience wrapper around :func:`recast_lon` with ``lon0 = -180``.

    Parameters
    ----------
    lon : np.ndarray
        An N-d array of longitudes in degrees

    Returns
    -------
    np.ndarray
        Converted longitudes in the range `[-180, 180[`

    Examples
    --------
    >>> recast_lon180(200)
    -160

    >>> recast_lon180(-200)
    160

    See Also
    --------
    :func:`recast_lon`, :func:`recast_lon360`
    """
    return recast_lon(lon, -180)


def plane_to_sphere(
    x: np.ndarray, y: np.ndarray, lon_origin: float = 0, lat_origin: float = 0
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert Cartesian coordinates on a plane to spherical coordinates.

    The arrays of input zonal and meridional displacements ``x`` and ``y`` are
    assumed to follow a contiguous trajectory. The spherical coordinate of each
    successive point is determined by following a great circle path from the
    previous point. The spherical coordinate of the first point is determined by
    following a great circle path from the origin, by default (0, 0).

    The output arrays have the same floating-point output type as the input.

    If projecting multiple trajectories onto the same plane, use
    :func:`apply_ragged` for highest accuracy.

    Parameters
    ----------
    x : np.ndarray
        An N-d array of zonal displacements in meters
    y : np.ndarray
        An N-d array of meridional displacements in meters
    lon_origin : float, optional
        Origin longitude of the tangent plane in degrees, default 0
    lat_origin : float, optional
        Origin latitude of the tangent plane in degrees, default 0

    Returns
    -------
    lon : np.ndarray
        Longitude in degrees
    lat : np.ndarray
        Latitude in degrees

    Examples
    --------
    >>> plane_to_sphere(np.array([0., 0.]), np.array([0., 1000.]))
    (array([0.00000000e+00, 5.50062664e-19]), array([0.       , 0.0089832]))

    You can also specify an origin longitude and latitude:

    >>> plane_to_sphere(np.array([0., 0.]), np.array([0., 1000.]), lon_origin=1, lat_origin=0)
    (array([1., 1.]), array([0.       , 0.0089832]))

    Raises
    ------
    AttributeError
        If ``x`` and ``y`` are not NumPy arrays

    See Also
    --------
    :func:`sphere_to_plane`
    """
    lon = np.empty_like(x)
    lat = np.empty_like(y)

    # Cartesian distances between each point
    dx = np.diff(x, prepend=0)
    dy = np.diff(y, prepend=0)

    distances = np.sqrt(dx**2 + dy**2)
    bearings = np.arctan2(dy, dx)

    # Compute spherical coordinates following great circles between each
    # successive point.
    lon[..., 0], lat[..., 0] = position_from_distance_and_bearing(
        lon_origin, lat_origin, distances[..., 0], bearings[..., 0]
    )
    for n in range(1, lon.shape[-1]):
        lon[..., n], lat[..., n] = position_from_distance_and_bearing(
            lon[..., n - 1], lat[..., n - 1], distances[..., n], bearings[..., n]
        )

    return lon, lat


def sphere_to_plane(
    lon: np.ndarray, lat: np.ndarray, lon_origin: float = 0, lat_origin: float = 0
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert spherical coordinates to a tangent (Cartesian) plane.

    The arrays of input longitudes and latitudes are assumed to be following
    a contiguous trajectory. The Cartesian coordinate of each successive point
    is determined by following a great circle path from the previous point.
    The Cartesian coordinate of the first point is determined by following a
    great circle path from the origin, by default (0, 0).

    The output arrays have the same floating-point output type as the input.

    If projecting multiple trajectories onto the same plane, use
    :func:`apply_ragged` for highest accuracy.

    Parameters
    ----------
    lon : np.ndarray
        An N-d array of longitudes in degrees
    lat : np.ndarray
        An N-d array of latitudes in degrees
    lon_origin : float, optional
        Origin longitude of the tangent plane in degrees, default 0
    lat_origin : float, optional
        Origin latitude of the tangent plane in degrees, default 0

    Returns
    -------
    x : np.ndarray
        x-coordinates on the tangent plane
    y : np.ndarray
        y-coordinates on the tangent plane

    Examples
    --------
    >>> sphere_to_plane(np.array([0., 1.]), np.array([0., 0.]))
    (array([     0.        , 111318.84502145]), array([0., 0.]))

    You can also specify an origin longitude and latitude:

    >>> sphere_to_plane(np.array([0., 1.]), np.array([0., 0.]), lon_origin=1, lat_origin=0)
    (array([-111318.84502145,       0.        ]),
     array([1.36326267e-11, 1.36326267e-11]))

    Raises
    ------
    AttributeError
        If ``lon`` and ``lat`` are not NumPy arrays

    See Also
    --------
    :func:`plane_to_sphere`
    """
    x = np.empty_like(lon)
    y = np.empty_like(lat)

    distances = np.empty_like(x)
    bearings = np.empty_like(x)

    # Distance and bearing of the starting point relative to the origin
    distances[0] = distance(lon_origin, lat_origin, lon[..., 0], lat[..., 0])
    bearings[0] = bearing(lon_origin, lat_origin, lon[..., 0], lat[..., 0])

    # Distance and bearing of the remaining points
    distances[1:] = distance(lon[..., :-1], lat[..., :-1], lon[..., 1:], lat[..., 1:])
    bearings[1:] = bearing(lon[..., :-1], lat[..., :-1], lon[..., 1:], lat[..., 1:])

    dx = distances * np.cos(bearings)
    dy = distances * np.sin(bearings)

    x[..., :] = np.cumsum(dx, axis=-1)
    y[..., :] = np.cumsum(dy, axis=-1)

    return x, y


def spherical_to_cartesian(
    lon: Union[float, list, np.ndarray, xr.DataArray],
    lat: Union[float, list, np.ndarray, xr.DataArray],
    radius: Optional[float] = EARTH_RADIUS_METERS,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Converts latitude and longitude on a spherical body to
     three-dimensional Cartesian coordinates.

    The Cartesian coordinate system is a right-handed system whose
    origin lies at the center of a sphere.  It is oriented with the
    Z-axis passing through the poles and the X-axis passing through
    the point lon = 0, lat = 0. This function is inverted by
    :func:`cartesian_to_spherical`.

    Parameters
    ----------
    lon : array-like
        An N-d array of longitudes in degrees.
    lat : array-like
        An N-d array of latitudes in degrees.
    radius: float, optional
        The radius of the spherical body in meters. The default assumes the Earth with
        EARTH_RADIUS_METERS = 6.3781e6.

    Returns
    -------
    x : float or array-like
        x-coordinates in 3D in meters.
    y : float or array-like
        y-coordinates in 3D in meters.
    z : float or array-like
        z-coordinates in 3D in meters.

    Examples
    --------
    >>> spherical_to_cartesian(np.array([0, 45]), np.array([0, 45]))
    (array([6378100., 3189050.]),
    array([      0., 3189050.]),
    array([      0.        , 4509997.76108592]))

    >>> spherical_to_cartesian(np.array([0, 45, 90]), np.array([0, 90, 180]), radius=1)
    (array([ 1.00000000e+00,  4.32978028e-17, -6.12323400e-17]),
    array([ 0.00000000e+00,  4.32978028e-17, -1.00000000e+00]),
    array([0.0000000e+00, 1.0000000e+00, 1.2246468e-16]))

    >>> x, y, z = spherical_to_cartesian(np.array([0, 5]), np.array([0, 5]))

    Raises
    ------
    AttributeError
        If ``lon`` and ``lat`` are not NumPy arrays.

    See Also
    --------
    :func:`cartesian_to_spherical`
    """
    lonr, latr = np.deg2rad(lon), np.deg2rad(lat)

    x = radius * np.cos(latr) * np.cos(lonr)
    y = radius * np.cos(latr) * np.sin(lonr)
    z = radius * np.sin(latr)

    return x, y, z


def cartesian_to_spherical(
    x: Union[float, np.ndarray, xr.DataArray],
    y: Union[float, np.ndarray, xr.DataArray],
    z: Union[float, np.ndarray, xr.DataArray],
) -> Tuple[np.ndarray, np.ndarray]:
    """Converts Cartesian three-dimensional coordinates to latitude and longitude on a
    spherical body.

    The Cartesian coordinate system is a right-handed system whose
    origin lies at the center of the sphere.  It is oriented with the
    Z-axis passing through the poles and the X-axis passing through
    the point lon = 0, lat = 0. This function is inverted by `spherical_to_cartesian`.

    Parameters
    ----------
    x : float or array-like
        x-coordinates in 3D.
    y : float or array-like
        y-coordinates in 3D.
    z : float or array-like
        z-coordinates in 3D.

    Returns
    -------
    lon : float or array-like
        An N-d array of longitudes in degrees in range [-180, 180].
    lat : float or array-like
        An N-d array of latitudes in degrees.

    Examples
    --------
    >>> x = EARTH_RADIUS_METERS * np.cos(np.deg2rad(45))
    >>> y = EARTH_RADIUS_METERS * np.cos(np.deg2rad(45))
    >>> z = 0 * x
    >>> cartesian_to_spherical(x, y, z)
    (44.99999999999985, 0.0)

    ``cartesian_to_spherical`` is inverted by ``spherical_to_cartesian``:

    >>> x, y, z = spherical_to_cartesian(np.array([45]),np.array(0))
    >>> cartesian_to_spherical(x, y, z)
    (array([45.]), array([0.]))

    Raises
    ------
    AttributeError
        If ``x``, ``y``, and ``z`` are not NumPy arrays.

    See Also
    --------
    :func:`spherical_to_cartesian`
    """

    R = np.sqrt(x**2 + y**2 + z**2)
    x /= R
    y /= R
    z /= R

    with np.errstate(divide="ignore"):
        lon = np.where(
            np.logical_and(x == 0, y == 0),
            0,
            recast_lon180(np.rad2deg(np.imag(np.log((x + 1j * y))))),
        )
    lat = np.rad2deg(np.arcsin(z))

    return lon, lat


def cartesian_to_tangentplane(
    u: Union[float, np.ndarray],
    v: Union[float, np.ndarray],
    w: Union[float, np.ndarray],
    longitude: Union[float, np.ndarray],
    latitude: Union[float, np.ndarray],
) -> Union[Tuple[float], Tuple[np.ndarray]]:
    """
    Project a three-dimensional Cartesian vector on a plane tangent to
    a spherical Earth.

    The Cartesian coordinate system is a right-handed system whose
    origin lies at the center of a sphere.  It is oriented with the
    Z-axis passing through the north pole at lat = 90, the X-axis passing through
    the point lon = 0, lat = 0, and the Y-axis passing through the point lon = 90,
    lat = 0.

    Parameters
    ----------
        u : float or np.ndarray
            First component of Cartesian vector.
        v : float or np.ndarray
            Second component of Cartesian vector.
        w : float or np.ndarray
            Third component of Cartesian vector.
        longitude : float or np.ndarray
            Longitude in degrees of tangent point of plane.
        latitude : float or np.ndarray
            Latitude in degrees of tangent point of plane.

    Returns
    -------
        up: float or np.ndarray
            First component of projected vector on tangent plane (positive eastward).
        vp: float or np.ndarray
            Second component of projected vector on tangent plane (positive northward).

    Raises
    ------
    Warning
        Raised if the input latitude is not in the expected range [-90, 90].

    Examples
    --------
    >>> u, v = cartesian_to_tangentplane(1, 1, 1, 45, 90)

    See Also
    --------
    :func:`tangentplane_to_cartesian`
    """
    if np.any(latitude < -90) or np.any(latitude > 90):
        warnings.warn("Input latitude outside of range [-90,90].")

    phi = np.radians(latitude)
    theta = np.radians(longitude)
    u_projected = v * np.cos(theta) - u * np.sin(theta)
    v_projected = (
        w * np.cos(phi)
        - u * np.cos(theta) * np.sin(phi)
        - v * np.sin(theta) * np.sin(phi)
    )
    # JML says vh = w.*cos(phi)-u.*cos(theta).*sin(phi)-v.*sin(theta).*sin(phi) but vh=w./cos(phi) is the same
    return u_projected, v_projected


def tangentplane_to_cartesian(
    up: Union[float, np.ndarray],
    vp: Union[float, np.ndarray],
    longitude: Union[float, np.ndarray],
    latitude: Union[float, np.ndarray],
) -> Union[Tuple[float], Tuple[np.ndarray]]:
    """
    Return the three-dimensional Cartesian components of a vector contained in
    a plane tangent to a spherical Earth.

    The Cartesian coordinate system is a right-handed system whose
    origin lies at the center of a sphere.  It is oriented with the
    Z-axis passing through the north pole at lat = 90, the X-axis passing through
    the point lon = 0, lat = 0, and the Y-axis passing through the point lon = 90,
    lat = 0.

    Parameters
    ----------
        up: float or np.ndarray
            First component of vector on tangent plane (positive eastward).
        vp: float or np.ndarray
            Second component of vector on tangent plane (positive northward).
        longitude : float or np.ndarray
            Longitude in degrees of tangent point of plane.
        latitude : float or np.ndarray
            Latitude in degrees of tangent point of plane.

    Returns
    -------
        u : float or np.ndarray
            First component of Cartesian vector.
        v : float or np.ndarray
            Second component of Cartesian vector.
        w : float or np.ndarray
            Third component of Cartesian vector.

    Examples
    --------
    >>> u, v, w = tangentplane_to_cartesian(1, 1, 45, 90)

    Notes
    -----
    This function is inverted by :func:`cartesian_to_tangetplane`.

    See Also
    --------
    :func:`cartesian_to_tangentplane`
    """
    phi = np.radians(latitude)
    theta = np.radians(longitude)
    u = -up * np.sin(theta) - vp * np.sin(phi) * np.cos(theta)
    v = up * np.cos(theta) - vp * np.sin(phi) * np.sin(theta)
    w = vp * np.cos(phi)

    return u, v, w


def coriolis_frequency(
    latitude: Union[float, np.ndarray],
) -> Union[float, np.ndarray]:
    """
    Return the Coriolis frequency or commonly known `f` parameter in geophysical fluid dynamics.

    Parameters
    ----------
    latitude : float or np.ndarray
        Latitude in degrees.

    Returns
    -------
    f : float or np.ndarray
        Signed Coriolis frequency in radian per seconds.

    Examples
    --------
    >>> f = coriolis_frequency(np.array([0, 45, 90]))
    """
    f = 2 * EARTH_ROTATION_RATE * np.sin(np.radians(latitude))

    return f
